#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2008-2009  Gary Burton
# Copyright (C) 2008       Robert Cheramy <robert@cheramy.net>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2012       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Export to GEDCOM"

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import os
import time

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.lib import (
    AttributeType,
    ChildRefType,
    Citation,
    Date,
    EventRoleType,
    EventType,
    LdsOrd,
    NameType,
    PlaceType,
    NoteType,
    Person,
    UrlType,
)
from gramps.version import VERSION
import gramps.plugins.lib.libgedcom as libgedcom
from gramps.gen.errors import DatabaseError

# keep the following line even though not obviously used (works on import)
from gramps.gui.plug.export import WriterOptionBox
from gramps.gen.updatecallback import UpdateCallback
from gramps.gen.utils.file import media_path_full
from gramps.gen.utils.place import conv_lat_lon
from gramps.gen.utils.location import get_main_location
from gramps.gen.display.place import displayer as _pd

# -------------------------------------------------------------------------
#
# GEDCOM tags representing attributes that may take a parameter, value or
# description on the same line as the tag
#
# -------------------------------------------------------------------------
NEEDS_PARAMETER = set(
    [
        "CAST",
        "DSCR",
        "EDUC",
        "IDNO",
        "NATI",
        "NCHI",
        "NMR",
        "OCCU",
        "PROP",
        "RELI",
        "SSN",
        "TITL",
    ]
)

LDS_ORD_NAME = {
    LdsOrd.BAPTISM: "BAPL",
    LdsOrd.ENDOWMENT: "ENDL",
    LdsOrd.SEAL_TO_PARENTS: "SLGC",
    LdsOrd.SEAL_TO_SPOUSE: "SLGS",
    LdsOrd.CONFIRMATION: "CONL",
}

LDS_STATUS = {
    LdsOrd.STATUS_BIC: "BIC",
    LdsOrd.STATUS_CANCELED: "CANCELED",
    LdsOrd.STATUS_CHILD: "CHILD",
    LdsOrd.STATUS_CLEARED: "CLEARED",
    LdsOrd.STATUS_COMPLETED: "COMPLETED",
    LdsOrd.STATUS_DNS: "DNS",
    LdsOrd.STATUS_INFANT: "INFANT",
    LdsOrd.STATUS_PRE_1970: "PRE-1970",
    LdsOrd.STATUS_QUALIFIED: "QUALIFIED",
    LdsOrd.STATUS_DNS_CAN: "DNS/CAN",
    LdsOrd.STATUS_STILLBORN: "STILLBORN",
    LdsOrd.STATUS_SUBMITTED: "SUBMITTED",
    LdsOrd.STATUS_UNCLEARED: "UNCLEARED",
}

LANGUAGES = {
    "cs": "Czech",
    "da": "Danish",
    "nl": "Dutch",
    "en": "English",
    "eo": "Esperanto",
    "fi": "Finnish",
    "fr": "French",
    "de": "German",
    "hu": "Hungarian",
    "it": "Italian",
    "lt": "Latvian",
    "lv": "Lithuanian",
    "no": "Norwegian",
    "po": "Polish",
    "pt": "Portuguese",
    "ro": "Romanian",
    "sk": "Slovak",
    "es": "Spanish",
    "sv": "Swedish",
    "ru": "Russian",
}

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------

QUALITY_MAP = {
    Citation.CONF_VERY_HIGH: "3",
    Citation.CONF_HIGH: "2",
    Citation.CONF_LOW: "1",
    Citation.CONF_VERY_LOW: "0",
}

PEDIGREE_TYPES = {
    ChildRefType.BIRTH: "birth",
    ChildRefType.STEPCHILD: "Step",
    ChildRefType.ADOPTED: "Adopted",
    ChildRefType.FOSTER: "Foster",
}

NOTES_PER_PERSON = 104  # fudge factor to make progress meter a bit smoother


# -------------------------------------------------------------------------
#
# sort_handles_by_id
#
# -------------------------------------------------------------------------
def sort_handles_by_id(handle_list, handle_to_object):
    """
    Sort a list of handles by the Gramps ID.

    The function that returns the object from the handle needs to be supplied
    so that we get the right object.

    """
    sorted_list = []
    for handle in handle_list:
        obj = handle_to_object(handle)
        if obj:
            data = (obj.get_gramps_id(), handle)
            sorted_list.append(data)
    sorted_list.sort()
    return sorted_list


# -------------------------------------------------------------------------
#
# breakup
#
# -------------------------------------------------------------------------
def breakup(txt, limit):
    """
    Break a line of text into a list of strings that conform to the
    maximum length specified, while breaking words in the middle of a word
    to avoid issues with spaces.
    """
    if limit < 1:
        raise ValueError("breakup: unexpected limit: %r" % limit)
    data = []
    while len(txt) > limit:
        # look for non-space pair to break between
        # fix issue #0012709 by removing Python2 code obsoleted by Python3
        idx = limit
        while idx > 0 and (txt[idx - 1].isspace() or txt[idx].isspace()):
            idx -= 1
        if idx == 0:
            # no words to break on, just break at limit anyway
            idx = limit
        data.append(txt[:idx])
        txt = txt[idx:]
    if len(txt) > 0:
        data.append(txt)
    return data


# -------------------------------------------------------------------------
#
# event_has_subordinate_data
#   may want to compare description w/ auto-generated one, and
#   if so, treat it same as if it were empty for this purpose
#
# -------------------------------------------------------------------------
def event_has_subordinate_data(event, event_ref):
    """determine if event is empty or not"""
    if event and event_ref:
        return (
            event.get_description().strip()
            or not event.get_date_object().is_empty()
            or event.get_place_handle()
            or event.get_attribute_list()
            or event_ref.get_attribute_list()
            or event.get_note_list()
            or event.get_citation_list()
            or event.get_media_list()
        )
    else:
        return False


# -------------------------------------------------------------------------
#
# GedcomWriter class
#
# -------------------------------------------------------------------------
class GedcomWriter(UpdateCallback):
    """
    The GEDCOM writer creates a GEDCOM file that contains the exported
    information from the database. It derives from UpdateCallback
    so that it can provide visual feedback via a progress bar if needed.
    """

    def __init__(self, database, user, option_box=None):
        UpdateCallback.__init__(self, user.callback)
        self.dbase = database
        self.dirname = None
        self.gedcom_file = None
        self.progress_cnt = 0
        self.setup(option_box)

    def setup(self, option_box):
        """
        If the option_box is present (GUI interface), then we check the
        "private", "restrict", and "cfilter" arguments to see if we need
        to apply proxy databases.
        """
        if option_box:
            option_box.parse_options()
            self.dbase = option_box.get_filtered_database(self.dbase, self)

    def write_gedcom_file(self, filename):
        """
        Write the actual GEDCOM file to the specified filename.
        """

        self.dirname = os.path.dirname(filename)
        with open(filename, "w", encoding="utf-8") as self.gedcom_file:
            person_len = self.dbase.get_number_of_people()
            family_len = self.dbase.get_number_of_families()
            source_len = self.dbase.get_number_of_sources()
            repo_len = self.dbase.get_number_of_repositories()
            note_len = self.dbase.get_number_of_notes() / NOTES_PER_PERSON

            total_steps = person_len + family_len + source_len + repo_len + note_len
            self.set_total(total_steps)
            self._header(filename)
            self._submitter()
            self._individuals()
            self._families()
            self._sources()
            self._repos()
            self._notes()
            self._all_media()

            self._writeln(0, "TRLR")

        return True

    def _writeln(self, level, token, textlines="", limit=72):
        """
        Write a line of text to the output file in the form of:

            LEVEL TOKEN text

        If the line contains newlines, it is broken into multiple lines using
        the CONT token. If any line is greater than the limit, it will broken
        into multiple lines using CONC.

        """
        assert token
        if textlines:
            # break the line into multiple lines if a newline is found
            textlines = textlines.replace("\n\r", "\n")
            textlines = textlines.replace("\r", "\n")
            # Need to double '@' See Gedcom 5.5 spec 'any_char'
            # but avoid xrefs and escapes
            if not textlines.startswith("@") and "@#" not in textlines:
                textlines = textlines.replace("@", "@@")
            textlist = textlines.split("\n")
            token_level = level
            for text in textlist:
                # make it unicode so that breakup below does the right thin.
                text = str(text)
                if limit:
                    prefix = "\n%d CONC " % (level + 1)
                    txt = prefix.join(breakup(text, limit))
                else:
                    txt = text
                self.gedcom_file.write("%d %s %s\n" % (token_level, token, txt))
                token_level = level + 1
                token = "CONT"
        else:
            self.gedcom_file.write("%d %s\n" % (level, token))

    def _header(self, filename):
        """
        Write the GEDCOM header.

            HEADER:=
            n HEAD {1:1}
            +1 SOUR <APPROVED_SYSTEM_ID> {1:1}
            +2 VERS <VERSION_NUMBER> {0:1}
            +2 NAME <NAME_OF_PRODUCT> {0:1}
            +2 CORP <NAME_OF_BUSINESS> {0:1}           # Not used
            +3 <<ADDRESS_STRUCTURE>> {0:1}             # Not used
            +2 DATA <NAME_OF_SOURCE_DATA> {0:1}        # Not used
            +3 DATE <PUBLICATION_DATE> {0:1}           # Not used
            +3 COPR <COPYRIGHT_SOURCE_DATA> {0:1}      # Not used
            +1 DEST <RECEIVING_SYSTEM_NAME> {0:1*}     # Not used
            +1 DATE <TRANSMISSION_DATE> {0:1}
            +2 TIME <TIME_VALUE> {0:1}
            +1 SUBM @XREF:SUBM@ {1:1}
            +1 SUBN @XREF:SUBN@ {0:1}
            +1 FILE <FILE_NAME> {0:1}
            +1 COPR <COPYRIGHT_GEDCOM_FILE> {0:1}
            +1 GEDC {1:1}
            +2 VERS <VERSION_NUMBER> {1:1}
            +2 FORM <GEDCOM_FORM> {1:1}
            +1 CHAR <CHARACTER_SET> {1:1}
            +2 VERS <VERSION_NUMBER> {0:1}
            +1 LANG <LANGUAGE_OF_TEXT> {0:1}
            +1 PLAC {0:1}
            +2 FORM <PLACE_HIERARCHY> {1:1}
            +1 NOTE <GEDCOM_CONTENT_DESCRIPTION> {0:1}
            +2 [CONT|CONC] <GEDCOM_CONTENT_DESCRIPTION> {0:M}

        """
        local_time = time.localtime(time.time())
        (year, mon, day, hour, minutes, sec) = local_time[0:6]
        date_str = "%d %s %d" % (day, libgedcom.MONTH[mon], year)
        time_str = "%02d:%02d:%02d" % (hour, minutes, sec)
        rname = self.dbase.get_researcher().get_name()

        self._writeln(0, "HEAD")
        self._writeln(1, "SOUR", "Gramps")
        self._writeln(2, "VERS", VERSION)
        self._writeln(2, "NAME", "Gramps")
        self._writeln(1, "DATE", date_str)
        self._writeln(2, "TIME", time_str)
        self._writeln(1, "SUBM", "@SUBM@")
        self._writeln(1, "FILE", filename, limit=255)
        self._writeln(1, "COPR", "Copyright (c) %d %s." % (year, rname))
        self._writeln(1, "GEDC")
        self._writeln(2, "VERS", "5.5.1")
        self._writeln(2, "FORM", "LINEAGE-LINKED")
        self._writeln(1, "CHAR", "UTF-8")

        # write the language string if the current LANG variable
        # matches something we know about.

        lang = glocale.language[0]
        if lang and len(lang) >= 2:
            lang_code = LANGUAGES.get(lang[0:2])
            if lang_code:
                self._writeln(1, "LANG", lang_code)

    def _submitter(self):
        """
        n @<XREF:SUBM>@ SUBM {1:1}
        +1 NAME <SUBMITTER_NAME> {1:1}
        +1 <<ADDRESS_STRUCTURE>> {0:1}
        +1 <<MULTIMEDIA_LINK>> {0:M}              # not used
        +1 LANG <LANGUAGE_PREFERENCE> {0:3}       # not used
        +1 RFN <SUBMITTER_REGISTERED_RFN> {0:1}   # not used
        +1 RIN <AUTOMATED_RECORD_ID> {0:1}        # not used
        +1 <<CHANGE_DATE>> {0:1}                  # not used
        """
        owner = self.dbase.get_researcher()
        name = owner.get_name()
        phon = owner.get_phone()
        mail = owner.get_email()

        self._writeln(0, "@SUBM@", "SUBM")
        self._writeln(1, "NAME", name)

        # Researcher is a sub-type of LocationBase, so get_city etc. which are
        # used in __write_addr work fine. However, the database owner street is
        # stored in address, so we need to temporarily copy it into street so
        # __write_addr works properly
        owner.set_street(owner.get_address())
        self.__write_addr(1, owner)

        if phon:
            self._writeln(1, "PHON", phon)
        if mail:
            self._writeln(1, "EMAIL", mail)

    def _individuals(self):
        """
        Write the individual people to the gedcom file.

        Since people like to have the list sorted by ID value, we need to go
        through a sorting step. We need to reset the progress bar, otherwise,
        people will be confused when the progress bar is idle.

        """
        self.set_text(_("Writing individuals"))
        phandles = self.dbase.iter_person_handles()

        sorted_list = []
        for handle in phandles:
            person = self.dbase.get_person_from_handle(handle)
            if person:
                data = (person.get_gramps_id(), handle)
                sorted_list.append(data)
        sorted_list.sort()

        for data in sorted_list:
            self.update()
            self._person(self.dbase.get_person_from_handle(data[1]))

    def _person(self, person):
        """
        Write out a single person.

        n @XREF:INDI@ INDI {1:1}
        +1 RESN <RESTRICTION_NOTICE> {0:1}            # not used
        +1 <<PERSONAL_NAME_STRUCTURE>> {0:M}
        +1 SEX <SEX_VALUE> {0:1}
        +1 <<INDIVIDUAL_EVENT_STRUCTURE>> {0:M}
        +1 <<INDIVIDUAL_ATTRIBUTE_STRUCTURE>> {0:M}
        +1 <<LDS_INDIVIDUAL_ORDINANCE>> {0:M}
        +1 <<CHILD_TO_FAMILY_LINK>> {0:M}
        +1 <<SPOUSE_TO_FAMILY_LINK>> {0:M}
        +1 SUBM @<XREF:SUBM>@ {0:M}
        +1 <<ASSOCIATION_STRUCTURE>> {0:M}
        +1 ALIA @<XREF:INDI>@ {0:M}
        +1 ANCI @<XREF:SUBM>@ {0:M}
        +1 DESI @<XREF:SUBM>@ {0:M}
        +1 <<SOURCE_CITATION>> {0:M}
        +1 <<MULTIMEDIA_LINK>> {0:M} ,*
        +1 <<NOTE_STRUCTURE>> {0:M}
        +1 RFN <PERMANENT_RECORD_FILE_NUMBER> {0:1}
        +1 AFN <ANCESTRAL_FILE_NUMBER> {0:1}
        +1 REFN <USER_REFERENCE_NUMBER> {0:M}
        +2 TYPE <USER_REFERENCE_TYPE> {0:1}
        +1 RIN <AUTOMATED_RECORD_ID> {0:1}
        +1 <<CHANGE_DATE>> {0:1}
        """
        if person is None:
            return
        self._writeln(0, "@%s@" % person.get_gramps_id(), "INDI")

        self._names(person)
        self._gender(person)
        self._person_event_ref("BIRT", person.get_birth_ref())
        self._person_event_ref("DEAT", person.get_death_ref())
        self._remaining_events(person)
        self._attributes(person)
        self._lds_ords(person, 1)
        self._child_families(person)
        self._parent_families(person)
        self._assoc(person, 1)
        self._person_sources(person)
        self._addresses(person)
        self._photos(person.get_media_list(), 1)
        self._url_list(person, 1)
        self._note_references(person.get_note_list(), 1)
        self._change(person.get_change_time(), 1)

    def _assoc(self, person, level):
        """
        n ASSO @<XREF:INDI>@ {0:M}
        +1 RELA <RELATION_IS_DESCRIPTOR> {1:1}
        +1 <<NOTE_STRUCTURE>> {0:M}
        +1 <<SOURCE_CITATION>> {0:M}
        """
        for ref in person.get_person_ref_list():
            person = self.dbase.get_person_from_handle(ref.ref)
            if person:
                self._writeln(level, "ASSO", "@%s@" % person.get_gramps_id())
                self._writeln(level + 1, "RELA", ref.get_relation())
                self._note_references(ref.get_note_list(), level + 1)
                self._source_references(ref.get_citation_list(), level + 1)

    def _note_references(self, notelist, level):
        """
        Write out the list of note handles to the current level.

        We use the Gramps ID as the XREF for the GEDCOM file.

        """
        for note_handle in notelist:
            note = self.dbase.get_note_from_handle(note_handle)
            if note:
                self._writeln(level, "NOTE", "@%s@" % note.get_gramps_id())

    def _names(self, person):
        """
        Write the names associated with the person to the current level.

        Since nicknames in version < 3.3 are separate from the name structure,
        we search the attribute list to see if we can find a nickname.
        Because we do not know the mappings, we just take the first nickname
        we find, and add it to the primary name.
        If a nickname is present in the name structure, it has precedence

        """
        nicknames = [
            attr.get_value()
            for attr in person.get_attribute_list()
            if int(attr.get_type()) == AttributeType.NICKNAME
        ]
        if len(nicknames) > 0:
            nickname = nicknames[0]
        else:
            nickname = ""

        self._person_name(person.get_primary_name(), nickname)
        for name in person.get_alternate_names():
            self._person_name(name, "")

    def _gender(self, person):
        """
        Write out the gender of the person to the file.

        If the gender is not male or female, simply do not output anything.
        The only valid values are M (male), F (female) and X (other). So if
        the gender is unknown, we output nothing.

        """
        if person.get_gender() == Person.MALE:
            self._writeln(1, "SEX", "M")
        elif person.get_gender() == Person.FEMALE:
            self._writeln(1, "SEX", "F")
        elif person.get_gender() == Person.OTHER:
            self._writeln(1, "SEX", "X")

    def _lds_ords(self, obj, level):
        """
        Simply loop through the list of LDS ordinances, and call the function
        that writes the LDS ordinance structure.
        """
        for lds_ord in obj.get_lds_ord_list():
            self.write_ord(lds_ord, level)

    def _remaining_events(self, person):
        """
        Output all events associated with the person that are not BIRTH or
        DEATH events.

        Because all we have are event references, we have to
        extract the real event to discover the event type.

        """
        global adop_written
        # adop_written is only shared between this function and
        # _process_person_event. This is rather ugly code, but it is difficult
        # to support an Adoption event without an Adopted relationship from the
        # parent(s), an Adopted relationship from the parent(s) without an
        # event, and both an event and a relationship. All these need to be
        # supported without duplicating the output of the ADOP GEDCOM tag. See
        # bug report 2370.
        adop_written = False
        for event_ref in person.get_event_ref_list():
            event = self.dbase.get_event_from_handle(event_ref.ref)
            if not event:
                continue
            self._process_person_event(person, event, event_ref)
        if not adop_written:
            self._adoption_records(person, adop_written)

    def _process_person_event(self, person, event, event_ref):
        """
        Process a person event, which is not a BIRTH or DEATH event.
        """
        global adop_written
        etype = int(event.get_type())
        # if the event is a birth or death, skip it.
        if etype in (EventType.BIRTH, EventType.DEATH):
            return

        role = int(event_ref.get_role())

        # if the event role is not primary, skip the event.
        if role != EventRoleType.PRIMARY:
            return

        val = libgedcom.PERSONALCONSTANTEVENTS.get(etype, "").strip()

        if val and val.strip():
            if val in NEEDS_PARAMETER:
                if event.get_description().strip():
                    self._writeln(1, val, event.get_description())
                else:
                    self._writeln(1, val)
            else:
                if event_has_subordinate_data(event, event_ref):
                    self._writeln(1, val)
                else:
                    self._writeln(1, val, "Y")
                if event.get_description().strip():
                    self._writeln(2, "TYPE", event.get_description())
        else:
            descr = event.get_description()
            if descr:
                self._writeln(1, "EVEN", descr)
            else:
                self._writeln(1, "EVEN")
            if val.strip():
                self._writeln(2, "TYPE", val)
            else:
                self._writeln(2, "TYPE", event.get_type().xml_str())
        self._dump_event_stats(event, event_ref)
        if etype == EventType.ADOPT and not adop_written:
            adop_written = True
            self._adoption_records(person, adop_written)

    def _adoption_records(self, person, adop_written):
        """
        Write Adoption events for each child that has been adopted.

        n ADOP
        +1 <<INDIVIDUAL_EVENT_DETAIL>>
        +1 FAMC @<XREF:FAM>@
        +2 ADOP <ADOPTED_BY_WHICH_PARENT>

        """

        adoptions = []

        for family in [
            self.dbase.get_family_from_handle(fh)
            for fh in person.get_parent_family_handle_list()
        ]:
            if family is None:
                continue
            for child_ref in [
                ref for ref in family.get_child_ref_list() if ref.ref == person.handle
            ]:
                if (
                    child_ref.mrel == ChildRefType.ADOPTED
                    or child_ref.frel == ChildRefType.ADOPTED
                ):
                    adoptions.append((family, child_ref.frel, child_ref.mrel))

        for fam, frel, mrel in adoptions:
            if not adop_written:
                self._writeln(1, "ADOP", "Y")
            self._writeln(2, "FAMC", "@%s@" % fam.get_gramps_id())
            if mrel == frel:
                self._writeln(3, "ADOP", "BOTH")
            elif mrel == ChildRefType.ADOPTED:
                self._writeln(3, "ADOP", "WIFE")
            else:
                self._writeln(3, "ADOP", "HUSB")

    def _attributes(self, person):
        """
        Write out the attributes to the GEDCOM file.

        Since we have already looked at nicknames when we generated the names,
        we filter them out here.

        We use the GEDCOM 5.5.1 FACT command to write out attributes not
        built in to GEDCOM.

        """

        # filter out the nicknames
        attr_list = [
            attr
            for attr in person.get_attribute_list()
            if attr.get_type() != AttributeType.NICKNAME
        ]

        for attr in attr_list:
            attr_type = int(attr.get_type())
            name = libgedcom.PERSONALCONSTANTATTRIBUTES.get(attr_type)
            key = attr.get_type().xml_str()
            value = attr.get_value().strip().replace("\r", " ")

            if key in ("AFN", "RFN", "REFN", "_UID", "_FSFTID"):
                self._writeln(1, key, value)
                continue

            if key == "RESN":
                self._writeln(1, "RESN")
                continue

            if name and name.strip():
                self._writeln(1, name, value)
            elif value:
                self._writeln(1, "FACT", value)
                self._writeln(2, "TYPE", key)
            else:
                continue
            self._note_references(attr.get_note_list(), 2)
            self._source_references(attr.get_citation_list(), 2)

    def _source_references(self, citation_list, level):
        """
        Loop through the list of citation handles, writing the information
        to the file.
        """
        for citation_handle in citation_list:
            self._source_ref_record(level, citation_handle)

    def _addresses(self, person):
        """
        Write out the addresses associated with the person as RESI events.
        """
        for addr in person.get_address_list():
            self._writeln(1, "RESI")
            self._date(2, addr.get_date_object())
            self.__write_addr(2, addr)
            if addr.get_phone():
                self._writeln(2, "PHON", addr.get_phone())

            self._note_references(addr.get_note_list(), 2)
            self._source_references(addr.get_citation_list(), 2)

    def _photos(self, media_list, level):
        """
        Loop through the list of media objects, writing the information
        to the file.
        """
        for photo in media_list:
            self._photo(photo, level)

    def _child_families(self, person):
        """
        Write the Gramps ID as the XREF for each family in which the person
        is listed as a child.
        """

        # get the list of familes from the handle list
        family_list = [
            self.dbase.get_family_from_handle(hndl)
            for hndl in person.get_parent_family_handle_list()
        ]

        for family in family_list:
            if family:
                self._writeln(1, "FAMC", "@%s@" % family.get_gramps_id())
                for child in family.get_child_ref_list():
                    if child.get_reference_handle() == person.get_handle():
                        if (
                            child.frel == ChildRefType.ADOPTED
                            and child.mrel == ChildRefType.ADOPTED
                        ):
                            self._writeln(2, "PEDI adopted")
                        elif (
                            child.frel == ChildRefType.BIRTH
                            and child.mrel == ChildRefType.BIRTH
                        ):
                            self._writeln(2, "PEDI birth")
                        elif (
                            child.frel == ChildRefType.STEPCHILD
                            and child.mrel == ChildRefType.STEPCHILD
                        ):
                            self._writeln(2, "PEDI stepchild")
                        elif (
                            child.frel == ChildRefType.FOSTER
                            and child.mrel == ChildRefType.FOSTER
                        ):
                            self._writeln(2, "PEDI foster")
                        elif child.frel == child.mrel:
                            self._writeln(2, "PEDI %s" % child.frel.xml_str())
                        else:
                            self._writeln(
                                2,
                                "_FREL %s"
                                % PEDIGREE_TYPES.get(
                                    child.frel.value, child.frel.xml_str()
                                ),
                            )
                            self._writeln(
                                2,
                                "_MREL %s"
                                % PEDIGREE_TYPES.get(
                                    child.mrel.value, child.mrel.xml_str()
                                ),
                            )

    def _parent_families(self, person):
        """
        Write the Gramps ID as the XREF for each family in which the person
        is listed as a parent.
        """

        # get the list of familes from the handle list
        family_list = [
            self.dbase.get_family_from_handle(hndl)
            for hndl in person.get_family_handle_list()
        ]

        for family in family_list:
            if family:
                self._writeln(1, "FAMS", "@%s@" % family.get_gramps_id())

    def _person_sources(self, person):
        """
        Loop through the list of citations, writing the information
        to the file.
        """
        for citation_handle in person.get_citation_list():
            self._source_ref_record(1, citation_handle)

    def _url_list(self, obj, level):
        """
        For Person's FAX, PHON, EMAIL, WWW lines;
        n PHON <PHONE_NUMBER> {0:3}
        n EMAIL <ADDRESS_EMAIL> {0:3}
        n FAX <ADDRESS_FAX> {0:3}
        n WWW <ADDRESS_WEB_PAGE> {0:3}

        n OBJE {1:1}
        +1 FORM <MULTIMEDIA_FORMAT> {1:1}
        +1 TITL <DESCRIPTIVE_TITLE> {0:1}
        +1 FILE <MULTIMEDIA_FILE_REFERENCE> {1:1}
        +1 <<NOTE_STRUCTURE>> {0:M}
        """
        for url in obj.get_url_list():
            if url.get_type() == UrlType.EMAIL:
                self._writeln(level, "EMAIL", url.get_path())
            elif url.get_type() == UrlType.WEB_HOME:
                self._writeln(level, "WWW", url.get_path())
            elif url.get_type() == _("Phone"):
                self._writeln(level, "PHON", url.get_path())
            elif url.get_type() == _("FAX"):
                self._writeln(level, "FAX", url.get_path())
            else:
                self._writeln(level, "OBJE")
                self._writeln(level + 1, "FORM", "URL")
                if url.get_description():
                    self._writeln(level + 1, "TITL", url.get_description())
                if url.get_path():
                    self._writeln(level + 1, "FILE", url.get_path(), limit=255)

    def _families(self):
        """
        Write out the list of families, sorting by Gramps ID.
        """
        self.set_text(_("Writing families"))
        # generate a list of (GRAMPS_ID, HANDLE) pairs. This list
        # can then be sorted by the sort routine, which will use the
        # first value of the tuple as the sort key.
        sorted_list = sort_handles_by_id(
            self.dbase.get_family_handles(), self.dbase.get_family_from_handle
        )

        # loop through the sorted list, pulling of the handle. This list
        # has already been sorted by GRAMPS_ID
        for family_handle in [hndl[1] for hndl in sorted_list]:
            self.update()
            self._family(self.dbase.get_family_from_handle(family_handle))

    def _family(self, family):
        """
        n @<XREF:FAM>@ FAM {1:1}
        +1 RESN <RESTRICTION_NOTICE> {0:1)
        +1 <<FAMILY_EVENT_STRUCTURE>> {0:M}
        +1 HUSB @<XREF:INDI>@ {0:1}
        +1 WIFE @<XREF:INDI>@ {0:1}
        +1 CHIL @<XREF:INDI>@ {0:M}
        +1 NCHI <COUNT_OF_CHILDREN> {0:1}
        +1 SUBM @<XREF:SUBM>@ {0:M}
        +1 <<LDS_SPOUSE_SEALING>> {0:M}
        +1 REFN <USER_REFERENCE_NUMBER> {0:M}
        """
        if family is None:
            return
        gramps_id = family.get_gramps_id()

        self._writeln(0, "@%s@" % gramps_id, "FAM")

        self._family_reference("HUSB", family.get_father_handle())
        self._family_reference("WIFE", family.get_mother_handle())

        self._lds_ords(family, 1)
        self._family_events(family)
        self._family_attributes(family.get_attribute_list(), 1)
        self._family_child_list(family.get_child_ref_list())
        self._source_references(family.get_citation_list(), 1)
        self._photos(family.get_media_list(), 1)
        self._note_references(family.get_note_list(), 1)
        self._change(family.get_change_time(), 1)

    def _family_child_list(self, child_ref_list):
        """
        Write the child XREF values to the GEDCOM file.
        """
        child_list = [
            self.dbase.get_person_from_handle(cref.ref).get_gramps_id()
            for cref in child_ref_list
        ]

        for gid in child_list:
            if gid is None:
                continue
            self._writeln(1, "CHIL", "@%s@" % gid)

    def _family_reference(self, token, person_handle):
        """
        Write the family reference to the file.

        This is either 'WIFE' or 'HUSB'. As usual, we use the Gramps ID as the
        XREF value.

        """
        if person_handle:
            person = self.dbase.get_person_from_handle(person_handle)
            if person:
                self._writeln(1, token, "@%s@" % person.get_gramps_id())

    def _family_events(self, family):
        """
        Output the events associated with the family.

        Because all we have are event references, we have to extract the real
        event to discover the event type.

        """
        for event_ref in family.get_event_ref_list():
            event = self.dbase.get_event_from_handle(event_ref.ref)
            if event is None:
                continue
            self._process_family_event(event, event_ref)
            self._dump_event_stats(event, event_ref)

    def _process_family_event(self, event, event_ref):
        """
        Process a single family event.
        """
        etype = int(event.get_type())
        val = libgedcom.FAMILYCONSTANTEVENTS.get(etype)

        if val:
            if event_has_subordinate_data(event, event_ref):
                self._writeln(1, val)
            else:
                self._writeln(1, val, "Y")

            if event.get_type() == EventType.MARRIAGE:
                self._family_event_attrs(event.get_attribute_list(), 2)

            if event.get_description().strip() != "":
                self._writeln(2, "TYPE", event.get_description())
        else:
            descr = event.get_description()
            if descr:
                self._writeln(1, "EVEN", descr)
            else:
                self._writeln(1, "EVEN")
            the_type = event.get_type()
            if the_type:
                self._writeln(2, "TYPE", the_type.xml_str())

    def _family_event_attrs(self, attr_list, level):
        """
        Write the attributes associated with the family event.

        The only ones we really care about are FATHER_AGE and MOTHER_AGE which
        we translate to WIFE/HUSB AGE attributes.

        """
        for attr in attr_list:
            if attr.get_type() == AttributeType.FATHER_AGE:
                self._writeln(level, "HUSB")
                self._writeln(level + 1, "AGE", attr.get_value())
            elif attr.get_type() == AttributeType.MOTHER_AGE:
                self._writeln(level, "WIFE")
                self._writeln(level + 1, "AGE", attr.get_value())

    def _family_attributes(self, attr_list, level):
        """
        Write out the attributes associated with a family to the GEDCOM file.

        Since we have already looked at nicknames when we generated the names,
        we filter them out here.

        We use the GEDCOM 5.5.1 FACT command to write out attributes not
        built in to GEDCOM.

        """

        for attr in attr_list:
            attr_type = int(attr.get_type())
            name = libgedcom.FAMILYCONSTANTATTRIBUTES.get(attr_type)
            key = attr.get_type().xml_str()
            value = attr.get_value().replace("\r", " ")

            if key in ("AFN", "RFN", "REFN", "_UID"):
                self._writeln(1, key, value)
                continue

            if name and name.strip():
                self._writeln(1, name, value)
                continue
            else:
                self._writeln(1, "FACT", value)
                self._writeln(2, "TYPE", key)

            self._note_references(attr.get_note_list(), level + 1)
            self._source_references(attr.get_citation_list(), level + 1)

    def _sources(self):
        """
        Write out the list of sources, sorting by Gramps ID.
        """
        self.set_text(_("Writing sources"))
        sorted_list = sort_handles_by_id(
            self.dbase.get_source_handles(), self.dbase.get_source_from_handle
        )

        for source_id, handle in sorted_list:
            self.update()
            source = self.dbase.get_source_from_handle(handle)
            if source is None:
                continue
            self._writeln(0, "@%s@" % source_id, "SOUR")
            if source.get_title():
                self._writeln(1, "TITL", source.get_title())

            if source.get_author():
                self._writeln(1, "AUTH", source.get_author())

            if source.get_publication_info():
                self._writeln(1, "PUBL", source.get_publication_info())

            if source.get_abbreviation():
                self._writeln(1, "ABBR", source.get_abbreviation())

            self._photos(source.get_media_list(), 1)

            for reporef in source.get_reporef_list():
                self._reporef(reporef, 1)
                # break

            self._note_references(source.get_note_list(), 1)
            self._change(source.get_change_time(), 1)

            for srcattr in source.get_attribute_list():
                if str(srcattr.type) == "_APID":
                    self._writeln(1, "_APID", srcattr.value)
                    break

    def _notes(self):
        """
        Write out the list of notes, sorting by Gramps ID.
        """
        self.set_text(_("Writing notes"))
        note_cnt = 0
        sorted_list = sort_handles_by_id(
            self.dbase.get_note_handles(), self.dbase.get_note_from_handle
        )

        for note_handle in [hndl[1] for hndl in sorted_list]:
            # the following makes the progress bar a bit smoother
            if not note_cnt % NOTES_PER_PERSON:
                self.update()
            note_cnt += 1
            note = self.dbase.get_note_from_handle(note_handle)
            if note is None:
                continue
            self._note_record(note)

    def _note_record(self, note):
        """
        n @<XREF:NOTE>@ NOTE <SUBMITTER_TEXT> {1:1}
        +1 [ CONC | CONT] <SUBMITTER_TEXT> {0:M}
        +1 <<SOURCE_CITATION>> {0:M}
        +1 REFN <USER_REFERENCE_NUMBER> {0:M}
        +2 TYPE <USER_REFERENCE_TYPE> {0:1}
        +1 RIN <AUTOMATED_RECORD_ID> {0:1}
        +1 <<CHANGE_DATE>> {0:1}
        """
        if note:
            self._writeln(0, "@%s@" % note.get_gramps_id(), "NOTE " + note.get())

    def _repos(self):
        """
        Write out the list of repositories, sorting by Gramps ID.

        REPOSITORY_RECORD:=
        n @<XREF:REPO>@ REPO {1:1}
        +1 NAME <NAME_OF_REPOSITORY> {1:1}
        +1 <<ADDRESS_STRUCTURE>> {0:1}
        +1 <<NOTE_STRUCTURE>> {0:M}
        +1 REFN <USER_REFERENCE_NUMBER> {0:M}
        +2 TYPE <USER_REFERENCE_TYPE> {0:1}
        +1 RIN <AUTOMATED_RECORD_ID> {0:1}
        +1 <<CHANGE_DATE>> {0:1}
        """
        self.set_text(_("Writing repositories"))
        sorted_list = sort_handles_by_id(
            self.dbase.get_repository_handles(), self.dbase.get_repository_from_handle
        )

        # GEDCOM only allows for a single repository per source

        for repo_id, handle in sorted_list:
            self.update()
            repo = self.dbase.get_repository_from_handle(handle)
            if repo is None:
                continue
            self._writeln(0, "@%s@" % repo_id, "REPO")
            if repo.get_name():
                self._writeln(1, "NAME", repo.get_name())
            for addr in repo.get_address_list():
                self.__write_addr(1, addr)
                if addr.get_phone():
                    self._writeln(1, "PHON", addr.get_phone())
            for url in repo.get_url_list():
                if url.get_type() == UrlType.EMAIL:
                    self._writeln(1, "EMAIL", url.get_path())
                elif url.get_type() == UrlType.WEB_HOME:
                    self._writeln(1, "WWW", url.get_path())
                elif url.get_type() == _("FAX"):
                    self._writeln(1, "FAX", url.get_path())
            self._note_references(repo.get_note_list(), 1)

    def _reporef(self, reporef, level):
        """
        n REPO [ @XREF:REPO@ | <NULL>] {1:1}
        +1 <<NOTE_STRUCTURE>> {0:M}
        +1 CALN <SOURCE_CALL_NUMBER> {0:M}
        +2 MEDI <SOURCE_MEDIA_TYPE> {0:1}
        """

        if reporef.ref is None:
            return

        repo = self.dbase.get_repository_from_handle(reporef.ref)
        if repo is None:
            return

        repo_id = repo.get_gramps_id()

        self._writeln(level, "REPO", "@%s@" % repo_id)

        self._note_references(reporef.get_note_list(), level + 1)

        if reporef.get_call_number():
            self._writeln(level + 1, "CALN", reporef.get_call_number())
            if reporef.get_media_type():
                self._writeln(level + 2, "MEDI", reporef.get_media_type().xml_str())

    def _person_event_ref(self, key, event_ref):
        """
        Write out the BIRTH and DEATH events for the person.
        """
        if event_ref:
            event = self.dbase.get_event_from_handle(event_ref.ref)
            if event_has_subordinate_data(event, event_ref):
                self._writeln(1, key)
            else:
                self._writeln(1, key, "Y")
            if event.get_description().strip() != "":
                self._writeln(2, "TYPE", event.get_description())
            self._dump_event_stats(event, event_ref)

    def _change(self, timeval, level):
        """
        CHANGE_DATE:=
            n CHAN {1:1}
            +1 DATE <CHANGE_DATE> {1:1}
            +2 TIME <TIME_VALUE> {0:1}
            +1 <<NOTE_STRUCTURE>>          # not used
        """
        self._writeln(level, "CHAN")
        time_val = time.localtime(timeval)
        self._writeln(
            level + 1,
            "DATE",
            "%d %s %d" % (time_val[2], libgedcom.MONTH[time_val[1]], time_val[0]),
        )
        self._writeln(
            level + 2,
            "TIME",
            "%02d:%02d:%02d" % (time_val[3], time_val[4], time_val[5]),
        )

    def _dump_event_stats(self, event, event_ref):
        """
        Write the event details for the event, using the event and event
        reference information.

        GEDCOM does not make a distinction between the two.

        """
        dateobj = event.get_date_object()
        self._date(2, dateobj)
        if self._datewritten:
            # write out TIME if present
            times = [
                attr.get_value()
                for attr in event.get_attribute_list()
                if int(attr.get_type()) == AttributeType.TIME
            ]
            # Not legal, but inserted by PhpGedView
            if len(times) > 0:
                self._writeln(3, "TIME", times[0])

        place = None

        if event.get_place_handle():
            place = self.dbase.get_place_from_handle(event.get_place_handle())
            self._place(place, dateobj, 2)

        for attr in event.get_attribute_list():
            attr_type = attr.get_type()
            if attr_type == AttributeType.CAUSE:
                self._writeln(2, "CAUS", attr.get_value())
            elif attr_type == AttributeType.AGENCY:
                self._writeln(2, "AGNC", attr.get_value())
            elif attr_type == _("Phone"):
                self._writeln(2, "PHON", attr.get_value())
            elif attr_type == _("FAX"):
                self._writeln(2, "FAX", attr.get_value())
            elif attr_type == _("EMAIL"):
                self._writeln(2, "EMAIL", attr.get_value())
            elif attr_type == _("WWW"):
                self._writeln(2, "WWW", attr.get_value())

        for attr in event_ref.get_attribute_list():
            attr_type = attr.get_type()
            if attr_type == AttributeType.AGE:
                self._writeln(2, "AGE", attr.get_value())
            elif attr_type == AttributeType.FATHER_AGE:
                self._writeln(2, "HUSB")
                self._writeln(3, "AGE", attr.get_value())
            elif attr_type == AttributeType.MOTHER_AGE:
                self._writeln(2, "WIFE")
                self._writeln(3, "AGE", attr.get_value())

        self._note_references(event.get_note_list(), 2)
        self._note_references(event_ref.get_note_list(), 2)

        self._source_references(event.get_citation_list(), 2)
        self._source_references(event_ref.get_citation_list(), 2)

        self._photos(event.get_media_list(), 2)
        if place:
            self._photos(place.get_media_list(), 2)

    def write_ord(self, lds_ord, index):
        """
        LDS_INDIVIDUAL_ORDINANCE:=
        [
           n [ BAPL | CONL ] {1:1}
          +1 DATE <DATE_LDS_ORD> {0:1}
          +1 TEMP <TEMPLE_CODE> {0:1}
          +1 PLAC <PLACE_LIVING_ORDINANCE> {0:1}
          +1 STAT <LDS_BAPTISM_DATE_STATUS> {0:1}
            +2 DATE <CHANGE_DATE> {1:1}
          +1 <<NOTE_STRUCTURE>> {0:M}
          +1 <<SOURCE_CITATION>> {0:M} p.39
        |
           n ENDL {1:1}
          +1 DATE <DATE_LDS_ORD> {0:1}
          +1 TEMP <TEMPLE_CODE> {0:1}
          +1 PLAC <PLACE_LIVING_ORDINANCE> {0:1}
          +1 STAT <LDS_ENDOWMENT_DATE_STATUS> {0:1}
            +2 DATE <CHANGE_DATE> {1:1}
          +1 <<NOTE_STRUCTURE>> {0:M}
          +1 <<SOURCE_CITATION>> {0:M}
        |
           n SLGC {1:1}
          +1 DATE <DATE_LDS_ORD> {0:1}
          +1 TEMP <TEMPLE_CODE> {0:1}
          +1 PLAC <PLACE_LIVING_ORDINANCE> {0:1}
          +1 FAMC @<XREF:FAM>@ {1:1}
          +1 STAT <LDS_CHILD_SEALING_DATE_STATUS> {0:1}
            +2 DATE <CHANGE_DATE> {1:1}
          +1 <<NOTE_STRUCTURE>> {0:M}
          +1 <<SOURCE_CITATION>> {0:M}
        ]
        """

        self._writeln(index, LDS_ORD_NAME[lds_ord.get_type()])
        self._date(index + 1, lds_ord.get_date_object())
        if lds_ord.get_family_handle():
            family_handle = lds_ord.get_family_handle()
            family = self.dbase.get_family_from_handle(family_handle)
            if family:
                self._writeln(index + 1, "FAMC", "@%s@" % family.get_gramps_id())
        if lds_ord.get_temple():
            self._writeln(index + 1, "TEMP", lds_ord.get_temple())
        if lds_ord.get_place_handle():
            place = self.dbase.get_place_from_handle(lds_ord.get_place_handle())
            self._place(place, lds_ord.get_date_object(), 2)
        if lds_ord.get_status() != LdsOrd.STATUS_NONE:
            self._writeln(2, "STAT", LDS_STATUS[lds_ord.get_status()])

        self._note_references(lds_ord.get_note_list(), index + 1)
        self._source_references(lds_ord.get_citation_list(), index + 1)

    def _date(self, level, date):
        """
        Write the 'DATE' GEDCOM token, along with the date in GEDCOM's
        expected format.
        """
        self._datewritten = True
        start = date.get_start_date()
        if start != Date.EMPTY:
            cal = date.get_calendar()
            mod = date.get_modifier()
            quality = None if mod else date.get_quality()
            if mod == Date.MOD_SPAN:
                val = "FROM %s TO %s" % (
                    libgedcom.make_gedcom_date(start, cal, mod, None),
                    libgedcom.make_gedcom_date(date.get_stop_date(), cal, mod, None),
                )
            elif mod == Date.MOD_RANGE:
                val = "BET %s AND %s" % (
                    libgedcom.make_gedcom_date(start, cal, mod, None),
                    libgedcom.make_gedcom_date(date.get_stop_date(), cal, mod, None),
                )
            else:
                val = libgedcom.make_gedcom_date(start, cal, mod, quality)
            self._writeln(level, "DATE", val)
        elif date.get_text():
            self._writeln(level, "DATE", date.get_text())
        else:
            self._datewritten = False

    def _person_name(self, name, attr_nick):
        """
        n NAME <NAME_PERSONAL> {1:1}
        +1 NPFX <NAME_PIECE_PREFIX> {0:1}
        +1 GIVN <NAME_PIECE_GIVEN> {0:1}
        +1 NICK <NAME_PIECE_NICKNAME> {0:1}
        +1 SPFX <NAME_PIECE_SURNAME_PREFIX {0:1}
        +1 SURN <NAME_PIECE_SURNAME> {0:1}
        +1 NSFX <NAME_PIECE_SUFFIX> {0:1}
        +1 <<SOURCE_CITATION>> {0:M}
        +1 <<NOTE_STRUCTURE>> {0:M}
        """
        gedcom_name = name.get_gedcom_name()

        firstname = name.get_first_name().strip()
        surns = []
        surprefix = ""
        for surn in name.get_surname_list():
            surns.append(surn.get_surname().replace("/", "?"))
            if surn.get_connector():
                # we store connector with the surname
                surns[-1] = surns[-1] + " " + surn.get_connector()
        surname = ", ".join(surns)
        if name.get_surname_list():
            # GEDCOM only supports a single surname prefix
            surn = name.get_surname_list()[0]
            surprefix = surn.get_prefix().replace("/", "?")
        suffix = name.get_suffix()
        title = name.get_title()
        nick = name.get_nick_name()
        call = name.get_call_name()
        if nick.strip() == "":
            nick = attr_nick

        self._writeln(1, "NAME", gedcom_name)
        if int(name.get_type()) == NameType.BIRTH:
            self._writeln(2, "TYPE", "birth")
        elif int(name.get_type()) == NameType.MARRIED:
            self._writeln(2, "TYPE", "married")
        elif int(name.get_type()) == NameType.AKA:
            self._writeln(2, "TYPE", "aka")
        else:
            self._writeln(2, "TYPE", name.get_type().xml_str())

        if firstname:
            self._writeln(2, "GIVN", firstname)
        if surprefix:
            self._writeln(2, "SPFX", surprefix)
        if surname:
            self._writeln(2, "SURN", surname)
        if call:
            self._writeln(2, "_RUFNAME", call)
        if name.get_suffix():
            self._writeln(2, "NSFX", suffix)
        if name.get_title():
            self._writeln(2, "NPFX", title)
        if nick:
            self._writeln(2, "NICK", nick)

        self._source_references(name.get_citation_list(), 2)
        self._note_references(name.get_note_list(), 2)

    def _source_ref_record(self, level, citation_handle):
        """
        n SOUR @<XREF:SOUR>@ /* pointer to source record */ {1:1}
        +1 PAGE <WHERE_WITHIN_SOURCE> {0:1}
        +1 EVEN <EVENT_TYPE_CITED_FROM> {0:1}
        +2 ROLE <ROLE_IN_EVENT> {0:1}
        +1 DATA {0:1}
        +2 DATE <ENTRY_RECORDING_DATE> {0:1}
        +2 TEXT <TEXT_FROM_SOURCE> {0:M}
        +3 [ CONC | CONT ] <TEXT_FROM_SOURCE> {0:M}
        +1 QUAY <CERTAINTY_ASSESSMENT> {0:1}
        +1 <<MULTIMEDIA_LINK>> {0:M} ,*
        +1 <<NOTE_STRUCTURE>> {0:M}
        """

        citation = self.dbase.get_citation_from_handle(citation_handle)
        if citation is None:  # removed by proxy
            return

        src_handle = citation.get_reference_handle()
        if src_handle is None:
            return

        src = self.dbase.get_source_from_handle(src_handle)
        if src is None:
            return

        # Reference to the source
        self._writeln(level, "SOUR", "@%s@" % src.get_gramps_id())
        if citation.get_page() != "":
            # PAGE <WHERE_WITHIN_SOURCE> can not have CONC lines.
            # WHERE_WITHIN_SOURCE:= {Size=1:248}
            # Maximize line to 248 and set limit to 248, for no line split
            self._writeln(level + 1, "PAGE", citation.get_page()[0:248], limit=248)

        conf = min(citation.get_confidence_level(), Citation.CONF_VERY_HIGH)
        if conf != Citation.CONF_NORMAL and conf != -1:
            self._writeln(level + 1, "QUAY", QUALITY_MAP[conf])

        if not citation.get_date_object().is_empty():
            self._writeln(level + 1, "DATA")
            self._date(level + 2, citation.get_date_object())

        if len(citation.get_note_list()) > 0:
            note_list = [
                self.dbase.get_note_from_handle(h) for h in citation.get_note_list()
            ]
            note_list = [n for n in note_list if n.get_type() == NoteType.SOURCE_TEXT]

            if note_list:
                ref_text = note_list[0].get()
            else:
                ref_text = ""

            if ref_text != "" and citation.get_date_object().is_empty():
                self._writeln(level + 1, "DATA")
            if ref_text != "":
                self._writeln(level + 2, "TEXT", ref_text)

            note_list = [
                self.dbase.get_note_from_handle(h) for h in citation.get_note_list()
            ]
            note_list = [
                n.handle
                for n in note_list
                if n and n.get_type() != NoteType.SOURCE_TEXT
            ]
            self._note_references(note_list, level + 1)

        self._photos(citation.get_media_list(), level + 1)

        even = None
        for srcattr in citation.get_attribute_list():
            if str(srcattr.type) == "EVEN":
                even = srcattr.value
                self._writeln(level + 1, "EVEN", even)
                break
        if even:
            for srcattr in citation.get_attribute_list():
                if str(srcattr.type) == "EVEN:ROLE":
                    self._writeln(level + 2, "ROLE", srcattr.value)
                    break

        for srcattr in citation.get_attribute_list():
            if str(srcattr.type) == "_APID":
                self._writeln(level + 1, "_APID", srcattr.value)
                break

    def _photo(self, photo, level):
        """
        n OBJE @<XREF:OBJE>@ {1:1}
        """
        photo_obj_id = photo.get_reference_handle()
        photo_obj = self.dbase.get_media_from_handle(photo_obj_id)
        if photo_obj:
            # if not os.path.isfile(path):
            # return
            self._writeln(level, "OBJE @%s@" % photo_obj.get_gramps_id())

    def _all_media(self):
        """
        Write out the list of media, sorting by Gramps ID.
        """
        self.set_text(_("Writing media"))
        # generate a list of (GRAMPS_ID, HANDLE) pairs. This list
        # can then be sorted by the sort routine, which will use the
        # first value of the tuple as the sort key.
        sorted_list = sort_handles_by_id(
            self.dbase.get_media_handles(), self.dbase.get_media_from_handle
        )

        # loop through the sorted list, pulling of the handle. This list
        # has already been sorted by GRAMPS_ID
        for media_handle in [hndl[1] for hndl in sorted_list]:
            self.update()
            self._media(self.dbase.get_media_from_handle(media_handle))

    def _media(self, media):
        """
        n @XREF:OBJE@ OBJE {1:1}
        +1 FILE <MULTIMEDIA_FILE_REFN> {1:M}
        +2 FORM <MULTIMEDIA_FORMAT> {1:1}
        +3 TYPE <SOURCE_MEDIA_TYPE> {0:1}
        +2 TITL <DESCRIPTIVE_TITLE> {0:1} p.48
        +1 REFN <USER_REFERENCE_NUMBER> {0:M}
        +2 TYPE <USER_REFERENCE_TYPE> {0:1}
        +1 RIN <AUTOMATED_RECORD_ID> {0:1}
        +1 <<NOTE_STRUCTURE>> {0:M}
        +1 <<SOURCE_CITATION>> {0:M}
        +1 <<CHANGE_DATE>> {0:1}
        """
        if media is None:
            return
        gramps_id = media.get_gramps_id()

        self._writeln(0, "@%s@" % gramps_id, "OBJE")
        form = os.path.splitext(media.get_path())[1][1:]
        path = media_path_full(self.dbase, media.get_path())
        self._writeln(1, "FILE", path, limit=255)
        if form:
            self._writeln(2, "FORM", form)
        self._writeln(2, "TITL", media.get_description())

        for attr in media.get_attribute_list():
            key = str(attr.get_type())
            value = attr.get_value().replace("\r", " ")
            if key in ("RIN", "RFN", "REFN"):
                self._writeln(1, key, value)
                continue
        self._note_references(media.get_note_list(), 1)
        self._source_references(media.get_citation_list(), 1)
        self._change(media.get_change_time(), 1)

    def _place(self, place, dateobj, level):
        """
        PLACE_STRUCTURE:=
            n PLAC <PLACE_NAME> {1:1}
            +1 FORM <PLACE_HIERARCHY> {0:1}
            +1 FONE <PLACE_PHONETIC_VARIATION> {0:M}  # not used
            +2 TYPE <PHONETIC_TYPE> {1:1}
            +1 ROMN <PLACE_ROMANIZED_VARIATION> {0:M} # not used
            +2 TYPE <ROMANIZED_TYPE> {1:1}
            +1 MAP {0:1}
            +2 LATI <PLACE_LATITUDE> {1:1}
            +2 LONG <PLACE_LONGITUDE> {1:1}
            +1 <<NOTE_STRUCTURE>> {0:M}
        """
        if place is None:
            return
        place_name = _pd.display(self.dbase, place, dateobj)
        self._writeln(level, "PLAC", place_name.replace("\r", " "), limit=120)
        longitude = place.get_longitude()
        latitude = place.get_latitude()
        if longitude and latitude:
            (latitude, longitude) = conv_lat_lon(latitude, longitude, "GEDCOM")
        if longitude and latitude:
            self._writeln(level + 1, "MAP")
            self._writeln(level + 2, "LATI", latitude)
            self._writeln(level + 2, "LONG", longitude)

        # The Gedcom standard shows that an optional address structure can
        # be written out in the event detail.
        # http://homepages.rootsweb.com/~pmcbride/gedcom/55gcch2.htm#EVENT_DETAIL
        location = get_main_location(self.dbase, place)
        street = location.get(PlaceType.STREET)
        locality = location.get(PlaceType.LOCALITY)
        city = location.get(PlaceType.CITY)
        state = location.get(PlaceType.STATE)
        country = location.get(PlaceType.COUNTRY)
        postal_code = place.get_code()

        if street or locality or city or state or postal_code or country:
            self._writeln(level, "ADDR", street)
            if street:
                self._writeln(level + 1, "ADR1", street)
            if locality:
                self._writeln(level + 1, "ADR2", locality)
            if city:
                self._writeln(level + 1, "CITY", city)
            if state:
                self._writeln(level + 1, "STAE", state)
            if postal_code:
                self._writeln(level + 1, "POST", postal_code)
            if country:
                self._writeln(level + 1, "CTRY", country)

        self._note_references(place.get_note_list(), level + 1)

    def __write_addr(self, level, addr):
        """
        n ADDR <ADDRESS_LINE> {0:1}
        +1 CONT <ADDRESS_LINE> {0:M}
        +1 ADR1 <ADDRESS_LINE1> {0:1}  (Street)
        +1 ADR2 <ADDRESS_LINE2> {0:1}  (Locality)
        +1 CITY <ADDRESS_CITY> {0:1}
        +1 STAE <ADDRESS_STATE> {0:1}
        +1 POST <ADDRESS_POSTAL_CODE> {0:1}
        +1 CTRY <ADDRESS_COUNTRY> {0:1}

        This is done along the lines suggested by Tamura Jones in
        http://www.tamurajones.net/GEDCOMADDR.xhtml as a result of bug 6382.
        "GEDCOM writers should always use the structured address format,
        and it use it for all addresses, including the submitter address and
        their own corporate address." "Vendors that want their product to pass
        even the strictest GEDCOM validation, should include export to the old
        free-form format..." [This goes on to say the free-form should be an
        option, but we have not made it an option in Gramps].

        @param level: The level number for the ADDR tag
        @type level: Integer
        @param addr: The location or address
        @type addr: [a super-type of] LocationBase
        """
        if (
            addr.get_street()
            or addr.get_locality()
            or addr.get_city()
            or addr.get_state()
            or addr.get_postal_code()
            or addr.get_country()
        ):
            self._writeln(level, "ADDR", addr.get_street())
            if addr.get_locality():
                self._writeln(level + 1, "CONT", addr.get_locality())
            if addr.get_city():
                self._writeln(level + 1, "CONT", addr.get_city())
            if addr.get_state():
                self._writeln(level + 1, "CONT", addr.get_state())
            if addr.get_postal_code():
                self._writeln(level + 1, "CONT", addr.get_postal_code())
            if addr.get_country():
                self._writeln(level + 1, "CONT", addr.get_country())

            if addr.get_street():
                self._writeln(level + 1, "ADR1", addr.get_street())
            if addr.get_locality():
                self._writeln(level + 1, "ADR2", addr.get_locality())
            if addr.get_city():
                self._writeln(level + 1, "CITY", addr.get_city())
            if addr.get_state():
                self._writeln(level + 1, "STAE", addr.get_state())
            if addr.get_postal_code():
                self._writeln(level + 1, "POST", addr.get_postal_code())
            if addr.get_country():
                self._writeln(level + 1, "CTRY", addr.get_country())


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def export_data(database, filename, user, option_box=None):
    """
    External interface used to register with the plugin system.
    """
    ret = False
    try:
        ged_write = GedcomWriter(database, user, option_box)
        ret = ged_write.write_gedcom_file(filename)
    except IOError as msg:
        msg2 = _("Could not create %s") % filename
        user.notify_error(msg2, str(msg))
    except DatabaseError as msg:
        user.notify_db_error("%s\n%s" % (_("GEDCOM Export failed"), str(msg)))
    return ret



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk

_LEFT = Gdk.keyval_from_name("Left")
_RIGHT = Gdk.keyval_from_name("Right")

# -------------------------------------------------------------------------
#
# Classes
#
# -------------------------------------------------------------------------


class GrampsTab(Gtk.Box):
    """
    This class provides the base level class for 'tabs', which are used to
    fill in notebook tabs for Gramps edit dialogs.

    Each tab returns a gtk container widget which can be inserted into a
    Gtk.Notebook by the instantiating object.

    All tab classes should inherit from GrampsTab
    """

    def __init__(self, dbstate, uistate, track, name):
        """
        @param dbstate: The database state. Contains a reference to
        the database, along with other state information. The GrampsTab
        uses this to access the database and to pass to and created
        child windows (such as edit dialogs).
        @type dbstate: DbState
        @param uistate: The UI state. Used primarily to pass to any created
        subwindows.
        @type uistate: DisplayState
        @param track: The window tracking mechanism used to manage windows.
        This is only used to pass to generted child windows.
        @type track: list
        @param name: Notebook label name
        @type name: str/unicode
        """
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        # store information to pass to child windows
        self.dbstate = dbstate
        self.uistate = uistate
        self.track = track
        self.changed = False
        self.__refs_for_deletion = []

        # save name used for notebook label, and build the widget used
        # for the label

        self.tab_name = name
        self.track_ref_for_deletion("tab_name")
        self.label_container = self.build_label_widget()
        self.track_ref_for_deletion("label_container")

        # build the interface
        self.share_btn = None
        self.build_interface()
        self.parent_notebook = None

    def get_selected(self):
        return None

    def is_empty(self):
        """
        Indicate if the tab contains any data. This is used to determine
        how the label should be displayed.
        """
        return True

    def build_label_widget(self):
        """
        Standard routine to build a widget. Does not need to be overridden
        by the derived class. Creates an container that has the label and
        the icon in it.
        @returns: widget to be used for the notebook label.
        @rtype: Gtk.Box
        """
        hbox = Gtk.Box()
        icon = self.get_icon_name()

        if isinstance(icon, tuple):
            name = icon[1]
        else:
            name = icon

        self.tab_image = Gtk.Image.new_from_icon_name(name, Gtk.IconSize.MENU)
        self.track_ref_for_deletion("tab_image")
        self.label = Gtk.Label(label=self.tab_name)
        self.track_ref_for_deletion("label")
        hbox.pack_start(self.tab_image, True, True, 0)
        hbox.set_spacing(6)
        hbox.add(self.label)
        hbox.show_all()
        return hbox

    def get_icon_name(self):
        """
        Provide the name of the registered stock icon to be used as the
        icon in the label. This is typically overridden by the derived
        class to provide the new name.
        @returns: stock icon name
        @rtype: str
        """
        return "document-new"

    def get_tab_widget(self):
        """
        Provide the widget to be used for the notebook tab label. A
        container class is provided, and the object may manipulate the
        child widgets contained in the container.
        @returns: gtk widget
        @rtype: Gtk.Box
        """
        return self.label_container

    def key_pressed(self, obj, event):
        """
        Handles the key being pressed.
        The inheriting object must contain a widget that connects at mimimum
        to this method, eg an eventbox, tree, ...
        """
        if event.type == Gdk.EventType.KEY_PRESS:
            if event.keyval in (_LEFT,) and (
                event.get_state() & Gdk.ModifierType.MOD1_MASK
            ):
                self.prev_page()
            elif event.keyval in (_RIGHT,) and (
                event.get_state() & Gdk.ModifierType.MOD1_MASK
            ):
                self.next_page()
            else:
                return
            return True

    def _set_label(self, show_image=True):
        """
        Updates the label based of if the tab contains information. Tabs
        without information will not have an icon, and the text will not
        be bold. Tabs that contain data will have their icon displayed and
        the label text will be in bold face.
        """
        if not self.is_empty():
            if show_image:
                self.tab_image.show()
            else:
                self.tab_image.hide()
            self.label.set_text("<b>%s</b>" % self.tab_name)
            self.label.set_use_markup(True)
        else:
            self.tab_image.hide()
            self.label.set_text(self.tab_name)
        self.label.set_use_underline(True)

    def build_interface(self):
        """
        Builds the interface for the derived class. This function should be
        overridden in the derived class. Since the classes are derived from
        Gtk.Box, the self.pack_start, self.pack_end, and self.add functions
        can be used to add widgets to the interface.
        """
        pass

    def set_parent_notebook(self, book):
        self.parent_notebook = book
        self.track_ref_for_deletion("parent_notebook")

    def next_page(self):
        if self.parent_notebook:
            self.parent_notebook.next_page()

    def prev_page(self):
        if self.parent_notebook:
            self.parent_notebook.prev_page()

    def track_ref_for_deletion(self, ref):
        """
        Record references of instance variables that need to be removed
        from scope so that the class can be garbage collected
        """
        if ref not in self.__refs_for_deletion:
            self.__refs_for_deletion.append(ref)

    def clean_up(self):
        """
        Remove any instance variables from scope which point to non-glade
        GTK objects so that the class can be garbage collected.
        """
        while len(self.__refs_for_deletion):
            attr = self.__refs_for_deletion.pop()
            delattr(self, attr)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Adapted from Graphviz.py (now deprecated)
#    Copyright (C) 2000-2007  Donald N. Allingham
#    Copyright (C) 2005-2006  Eero Tamminen
#    Copyright (C) 2007-2008  Brian G. Matherly
#    Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
#    Contributions by Lorenzo Cappelletti <lorenzo.cappelletti@email.it>
#    Copyright (C) 2008       Stephane Charette <stephanecharette@gmail.com>
#    Copyright (C) 2009       Gary Burton
#    Contribution 2009 by     Bob Ham <rah@bash.sh>
#    Copyright (C) 2010       Jakim Friant
#    Copyright (C) 2013       Fedir Zinchuk <fedikw@gmail.com>
#    Copyright (C) 2013-2015  Paul Franklin
#    Copyright (C) 2015       Fabrice <fobrice@laposte.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Create a relationship graph using Graphviz
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from functools import partial
import html

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.plug.menu import (
    BooleanOption,
    EnumeratedListOption,
    FilterOption,
    PersonOption,
    ColorOption,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.lib import ChildRefType, EventRoleType, EventType, Date
from gramps.gen.utils.file import media_path_full, find_file
from gramps.gen.utils.thumbnails import get_thumbnail_path
from gramps.gen.relationship import get_relationship_calculator
from gramps.gen.utils.db import (
    get_birth_or_fallback,
    get_death_or_fallback,
    find_parents,
)
from gramps.gen.display.place import displayer as _pd
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.errors import ReportError

# ------------------------------------------------------------------------
#
# Constant options items
#
# ------------------------------------------------------------------------
_COLORS = [
    {"name": _("B&W outline"), "value": "outlined"},
    {"name": _("Colored outline"), "value": "colored"},
    {"name": _("Color fill"), "value": "filled"},
]

_ARROWS = [
    {"name": _("Descendants <- Ancestors"), "value": "d"},
    {"name": _("Descendants -> Ancestors"), "value": "a"},
    {"name": _("Descendants <-> Ancestors"), "value": "da"},
    {"name": _("Descendants - Ancestors"), "value": ""},
]


# ------------------------------------------------------------------------
#
# RelGraphReport class
#
# ------------------------------------------------------------------------
class RelGraphReport(Report):
    def __init__(self, database, options, user):
        """
        Create RelGraphReport object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        filter     - Filter to be applied to the people of the database.
                     The option class carries its number, and the function
                     returning the list of filters.
        arrow      - Arrow styles for heads and tails.
        showfamily - Whether to show family nodes.
        inc_id     - Whether to include IDs.
        url        - Whether to include URLs.
        inclimg    - Include images or not
        imgpos     - Image position, above/beside name
        color      - Whether to use outline, colored outline or filled color
                     in graph
        color_males    - Colour to apply to males
        color_females  - Colour to apply to females
        color_unknown  - Colour to apply to unknown genders
        color_families - Colour to apply to families
        dashed         - Whether to use dashed lines for non-birth relationships
        use_roundedcorners - Whether to use rounded corners for females
        use_hexagons   - Whether to use hexagons for individuals of unknown gender
        name_format    - Preferred format to display names
        incl_private   - Whether to include private data
        event_choice   - Whether to include dates and/or places
        occupation     - Whether to include occupations
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """
        Report.__init__(self, database, options, user)

        menu = options.menu
        get_option_by_name = options.menu.get_option_by_name
        get_value = lambda name: get_option_by_name(name).get_value()

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)
        self._db = self.database

        self.includeid = get_value("inc_id")
        self.includeurl = get_value("url")
        self.includeimg = get_value("includeImages")
        self.imgpos = get_value("imageOnTheSide")
        self.use_roundedcorners = get_value("useroundedcorners")
        self.use_hexagons = get_value("usehexagons")
        self.adoptionsdashed = get_value("dashed")
        self.show_families = get_value("showfamily")
        self.show_family_leaves = get_value("show_family_leaves")
        self.use_subgraphs = get_value("usesubgraphs")
        self.event_choice = get_value("event_choice")
        self.occupation = get_value("occupation")
        self.use_html_output = False

        self.colorize = get_value("color")
        color_males = get_value("colormales")
        color_females = get_value("colorfemales")
        color_other = get_value("colorother")
        color_unknown = get_value("colorunknown")
        color_families = get_value("colorfamilies")
        self.colors = {
            "male": color_males,
            "female": color_females,
            "other": color_other,
            "unknown": color_unknown,
            "family": color_families,
        }

        arrow_str = get_value("arrow")
        if "d" in arrow_str:
            self.arrowheadstyle = "normal"
        else:
            self.arrowheadstyle = "none"
        if "a" in arrow_str:
            self.arrowtailstyle = "normal"
        else:
            self.arrowtailstyle = "none"
        filter_option = get_option_by_name("filter")
        self._filter = filter_option.get_filter()

        stdoptions.run_name_format_option(self, menu)

        pid = get_value("pid")
        self.center_person = self._db.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        self.increlname = get_value("increlname")
        if self.increlname:
            self.rel_calc = get_relationship_calculator(
                reinit=True, clocale=self._locale
            )

        if __debug__:
            self.advrelinfo = get_value("advrelinfo")
        else:
            self.advrelinfo = False

    def write_report(self):
        person_handles = self._filter.apply(
            self._db, self._db.iter_person_handles(), user=self._user
        )
        # Hash people in a dictionary for faster inclusion checking
        self.persons = set(person_handles)

        person_handles = self.sort_persons(person_handles)

        if len(person_handles) > 1:
            if self._user:
                self._user.begin_progress(
                    _("Relationship Graph"),
                    _("Generating report"),
                    len(person_handles) * 2,
                )
            self.add_persons_and_families(person_handles)
            self.add_child_links_to_families(person_handles)
            if self._user:
                self._user.end_progress()

    def sort_persons(self, person_handle_list):
        "sort persons by close relations"

        # first make a list of all persons who don't have any parents
        root_nodes = list()
        for person_handle in person_handle_list:
            person = self.database.get_person_from_handle(person_handle)
            has_parent = False
            for parent_handle in find_parents(self.database, person):
                if parent_handle not in self.persons:
                    continue
                has_parent = True
            if not has_parent:
                root_nodes.append(person_handle)

        # now start from all root nodes we found and traverse their trees
        outlist = list()
        p_done = set()
        for person_handle in root_nodes:
            todolist = list()
            todolist.append(person_handle)
            while len(todolist) > 0:
                # take the first person from todolist and do sanity check
                cur = todolist.pop(0)
                if cur in p_done:
                    continue
                if cur not in self.persons:
                    p_done.add(cur)
                    continue
                person = self.database.get_person_from_handle(cur)

                # first check whether both parents are added
                missing_parents = False
                for parent_handle in find_parents(self.database, person):
                    if not parent_handle or parent_handle in p_done:
                        continue
                    if parent_handle not in self.persons:
                        continue
                    todolist.insert(0, parent_handle)
                    missing_parents = True

                # if one of the parents is still missing, wait for them
                if missing_parents:
                    continue

                # add person to the sorted output
                outlist.append(cur)
                p_done.add(cur)

                # add all spouses and children to the todo list
                family_list = person.get_family_handle_list()
                for fam_handle in family_list:
                    family = self.database.get_family_from_handle(fam_handle)
                    if family is None:
                        continue
                    if family.get_father_handle() and family.get_father_handle() != cur:
                        todolist.insert(0, family.get_father_handle())
                    if family.get_mother_handle() and family.get_mother_handle() != cur:
                        todolist.insert(0, family.get_mother_handle())
                    for child_ref in family.get_child_ref_list():
                        todolist.append(child_ref.ref)

        # finally store the result
        assert len(person_handle_list) == len(outlist)
        return outlist

    def add_child_links_to_families(self, person_handles):
        """
        returns string of Graphviz edges linking parents to families or
        children
        """
        for person_handle in person_handles:
            if self._user:
                self._user.step_progress()
            person = self._db.get_person_from_handle(person_handle)
            p_id = person.get_gramps_id()
            for fam_handle in person.get_parent_family_handle_list():
                family = self._db.get_family_from_handle(fam_handle)
                father_handle = family.get_father_handle()
                mother_handle = family.get_mother_handle()
                sibling = False
                for child_ref in family.get_child_ref_list():
                    if child_ref.ref == person_handle:
                        frel = child_ref.frel
                        mrel = child_ref.mrel
                    elif child_ref.ref in self.persons:
                        sibling = True
                if self.show_families and (
                    (father_handle and father_handle in self.persons)
                    or (mother_handle and mother_handle in self.persons)
                    or sibling
                ):
                    # Link to the family node if either parent is in graph
                    self.add_family_link(p_id, family, frel, mrel)
                else:
                    # Link to the parents' nodes directly, if they are in graph
                    if father_handle and father_handle in self.persons:
                        self.add_parent_link(p_id, father_handle, frel)
                    if mother_handle and mother_handle in self.persons:
                        self.add_parent_link(p_id, mother_handle, mrel)

    def add_family_link(self, p_id, family, frel, mrel):
        "Links the child to a family"
        style = "solid"
        adopted = (int(frel) != ChildRefType.BIRTH) or (int(mrel) != ChildRefType.BIRTH)
        # If birth relation to father is NONE, meaning there is no father and
        # if birth relation to mother is BIRTH then solid line
        if int(frel) == ChildRefType.NONE and int(mrel) == ChildRefType.BIRTH:
            adopted = False
        if adopted and self.adoptionsdashed:
            style = "dotted"
        self.doc.add_link(
            family.get_gramps_id(),
            p_id,
            style,
            self.arrowheadstyle,
            self.arrowtailstyle,
        )

    def add_parent_link(self, p_id, parent_handle, rel):
        "Links the child to a parent"
        style = "solid"
        if (int(rel) != ChildRefType.BIRTH) and self.adoptionsdashed:
            style = "dotted"
        parent = self._db.get_person_from_handle(parent_handle)
        self.doc.add_link(
            parent.get_gramps_id(),
            p_id,
            style,
            self.arrowheadstyle,
            self.arrowtailstyle,
        )

    def add_persons_and_families(self, person_handles):
        "adds nodes for persons and their families"
        # variable to communicate with get_person_label
        self.use_html_output = False

        # The list of families for which we have output the node,
        # so we don't do it twice
        families_done = set()
        for person_handle in person_handles:
            if self._user:
                self._user.step_progress()
            # determine per person if we use HTML style label
            if self.includeimg:
                self.use_html_output = True
            person = self._db.get_person_from_handle(person_handle)
            if person is None:
                continue
            p_id = person.get_gramps_id()
            # Output the person's node
            label = self.get_person_label(person)
            (shape, style, color, fill) = self.get_gender_style(person)
            url = ""
            if self.includeurl:
                phan = person_handle
                dirpath = "ppl/%s/%s" % (phan[-1], phan[-2])
                dirpath = dirpath.lower()
                url = "%s/%s.html" % (dirpath, phan)

            self.doc.add_node(p_id, label, shape, color, style, fill, url)

            # Output families where person is a parent
            if self.show_families:
                family_list = person.get_family_handle_list()
                for fam_handle in family_list:
                    family = self._db.get_family_from_handle(fam_handle)
                    if family is None:
                        continue
                    if fam_handle not in families_done:
                        if not self.show_family_leaves:
                            family_members = {
                                family.father_handle,
                                family.mother_handle,
                            }.union(
                                child_ref.ref for child_ref in family.child_ref_list
                            ) - {
                                None
                            }
                            if len(family_members.intersection(person_handles)) < 2:
                                continue
                        self.__add_family(fam_handle)
                        families_done.add(fam_handle)
                    # If subgraphs are not chosen then each parent is linked
                    # separately to the family. This gives Graphviz greater
                    # control over the layout of the whole graph but
                    # may leave spouses not positioned together.
                    if not self.use_subgraphs and fam_handle in families_done:
                        self.doc.add_link(
                            p_id,
                            family.get_gramps_id(),
                            "",
                            self.arrowheadstyle,
                            self.arrowtailstyle,
                        )

                # Output families where person is a sibling if another sibling
                # is present
                family_list = person.get_parent_family_handle_list()
                for fam_handle in family_list:
                    if fam_handle in families_done:
                        continue
                    family = self.database.get_family_from_handle(fam_handle)
                    if family is None:
                        continue
                    for child_ref in family.get_child_ref_list():
                        if (
                            child_ref.ref != person_handle
                            and child_ref.ref in self.persons
                        ):
                            families_done.add(fam_handle)
                            self.__add_family(fam_handle)

    def __add_family(self, fam_handle):
        """Add a node for a family and optionally link the spouses to it"""
        fam = self._db.get_family_from_handle(fam_handle)
        if fam is None:
            return
        fam_id = fam.get_gramps_id()

        m_type = m_date = m_place = ""
        d_type = d_date = d_place = ""
        for event_ref in fam.get_event_ref_list():
            event = self._db.get_event_from_handle(event_ref.ref)
            if event is None:
                continue
            if event.type == EventType.MARRIAGE and (
                event_ref.get_role() == EventRoleType.FAMILY
                or event_ref.get_role() == EventRoleType.PRIMARY
            ):
                m_type = event.type
                m_date = self.get_date_string(event)
                if not (self.event_choice == 3 and m_date):
                    m_place = self.get_place_string(event)
                break
            if event.type == EventType.DIVORCE and (
                event_ref.get_role() == EventRoleType.FAMILY
                or event_ref.get_role() == EventRoleType.PRIMARY
            ):
                d_type = event.type
                d_date = self.get_date_string(event)
                if not (self.event_choice == 3 and d_date):
                    d_place = self.get_place_string(event)
                break

        labellines = list()
        if self.includeid == 2:
            # id on separate line
            labellines.append("(%s)" % fam_id)
        if self.event_choice == 7:
            if m_type:
                line = m_type.get_abbreviation()
                if m_date:
                    line += " %s" % m_date
                if m_date and m_place:
                    labellines.append(line)
                    line = ""
                if m_place:
                    line += " %s" % m_place
                labellines.append(line)
            if d_type:
                line = d_type.get_abbreviation()
                if d_date:
                    line += " %s" % d_date
                if d_date and d_place:
                    labellines.append(line)
                    line = ""
                if d_place:
                    line += " %s" % d_place
                labellines.append(line)
        else:
            if m_date:
                labellines.append("(%s)" % m_date)
            if m_place:
                labellines.append("(%s)" % m_place)

        label = "\\n".join(labellines)
        labellines = list()
        if self.includeid == 1:
            # id on same line
            labellines.append("(%s)" % fam_id)
        if len(label):
            labellines.append(label)
        label = " ".join(labellines)

        color = ""
        fill = ""
        style = "solid"
        if self.colorize == "colored":
            color = self.colors["family"]
        elif self.colorize == "filled":
            fill = self.colors["family"]
            style = "filled"
        self.doc.add_node(fam_id, label, "ellipse", color, style, fill)

        # If subgraphs are used then we add both spouses here and Graphviz
        # will attempt to position both spouses closely together.
        # TODO: A person who is a parent in more than one family may only be
        #       positioned next to one of their spouses. The code currently
        #       does not take into account multiple spouses.
        if self.use_subgraphs:
            self.doc.start_subgraph(fam_id)
            f_handle = fam.get_father_handle()
            m_handle = fam.get_mother_handle()
            if (
                self.use_subgraphs == 2
                and f_handle in self.persons
                and m_handle in self.persons
            ):
                father = self._db.get_person_from_handle(f_handle)
                mother = self._db.get_person_from_handle(m_handle)
                fcount = 0
                mcount = 0
                family_list = father.get_parent_family_handle_list()
                for fam_handle in family_list:
                    family = self._db.get_family_from_handle(fam_handle)
                    if family is not None:
                        fcount = fcount + 1
                family_list = mother.get_parent_family_handle_list()
                for fam_handle in family_list:
                    family = self._db.get_family_from_handle(fam_handle)
                    if family is not None:
                        mcount = mcount + 1
                first = father
                second = mother
                if fcount < mcount:
                    first = mother
                    second = father
                self.doc.add_link(
                    first.get_gramps_id(),
                    second.get_gramps_id(),
                    "invis",
                    "none",
                    "none",
                )
            if f_handle in self.persons:
                father = self._db.get_person_from_handle(f_handle)
                self.doc.add_link(
                    father.get_gramps_id(),
                    fam_id,
                    "",
                    self.arrowheadstyle,
                    self.arrowtailstyle,
                )
            if m_handle in self.persons:
                mother = self._db.get_person_from_handle(m_handle)
                self.doc.add_link(
                    mother.get_gramps_id(),
                    fam_id,
                    "",
                    self.arrowheadstyle,
                    self.arrowtailstyle,
                )
            self.doc.end_subgraph()

    def get_gender_style(self, person):
        "return gender specific person style"
        gender = person.get_gender()
        shape = "box"
        style = "solid"
        color = ""
        fill = ""

        if gender == person.FEMALE and self.use_roundedcorners:
            style = "rounded"
        elif gender == person.UNKNOWN and self.use_hexagons:
            shape = "hexagon"

        if person == self.center_person and self.increlname:
            shape = "octagon"

        if self.colorize == "colored":
            if gender == person.MALE:
                color = self.colors["male"]
            elif gender == person.FEMALE:
                color = self.colors["female"]
            elif gender == person.OTHER:
                color = self.colors["other"]
            else:
                color = self.colors["unknown"]
        elif self.colorize == "filled":
            style += ",filled"
            if gender == person.MALE:
                fill = self.colors["male"]
            elif gender == person.FEMALE:
                fill = self.colors["female"]
            elif gender == person.OTHER:
                fill = self.colors["other"]
            else:
                fill = self.colors["unknown"]
        return (shape, style, color, fill)

    def get_person_label(self, person):
        "return person label string"
        # see if we have an image to use for this person
        image_path = None
        if self.use_html_output:
            media_list = person.get_media_list()
            if len(media_list) > 0:
                media_handle = media_list[0].get_reference_handle()
                media = self._db.get_media_from_handle(media_handle)
                media_mime_type = media.get_mime_type()
                if media_mime_type[0:5] == "image":
                    image_path = get_thumbnail_path(
                        media_path_full(self._db, media.get_path()),
                        rectangle=media_list[0].get_rectangle(),
                    )
                    # test if thumbnail actually exists in thumbs
                    # (import of data means media files might not be present
                    image_path = find_file(image_path)

        label = ""
        line_delimiter = "\\n"

        # If we have an image, then start an HTML table; remember to close
        # the table afterwards!
        #
        # This isn't a free-form HTML format here...just a few keywords that
        # happen to be
        # similar to keywords commonly seen in HTML.  For additional
        # information on what
        # is allowed, see:
        #
        #       http://www.graphviz.org/info/shapes.html#html
        #
        if self.use_html_output and image_path:
            line_delimiter = "<BR/>"
            label += '<TABLE BORDER="0" CELLSPACING="2" CELLPADDING="0" '
            label += 'CELLBORDER="0"><TR><TD></TD><TD>'
            label += '<IMG SRC="%s"/></TD><TD></TD>' % image_path
            if self.imgpos == 0:
                # trick it into not stretching the image
                label += '</TR><TR><TD COLSPAN="3">'
            else:
                label += "<TD>"
        else:
            # no need for html label with this person
            self.use_html_output = False

        # at the very least, the label must have the person's name
        p_name = self._name_display.display(person)
        label += html.escape(p_name)
        p_id = person.get_gramps_id()
        if self.includeid == 1:  # same line
            label += " (%s)" % p_id
        elif self.includeid == 2:  # own line
            label += "%s(%s)" % (line_delimiter, p_id)
        if self.event_choice != 0:
            b_date, d_date, b_place, d_place, b_type, d_type = self.get_event_strings(
                person
            )
            if self.event_choice in [1, 2, 3, 4, 5] and (b_date or d_date):
                label += "%s(" % line_delimiter
                if b_date:
                    label += "%s" % b_date
                label += " – "
                if d_date:
                    label += "%s" % d_date
                label += ")"
            if (
                self.event_choice in [2, 3, 5, 6]
                and (b_place or d_place)
                and not (self.event_choice == 3 and (b_date or d_date))
            ):
                label += "%s(" % line_delimiter
                if b_place:
                    label += "%s" % b_place
                label += " – "
                if d_place:
                    label += "%s" % d_place
                label += ")"
        if self.event_choice == 7:
            if b_type:
                label += "%s%s" % (line_delimiter, b_type.get_abbreviation())
                if b_date:
                    label += " %s" % b_date
                if b_place:
                    label += " %s" % b_place

            if d_type:
                label += "%s%s" % (line_delimiter, d_type.get_abbreviation())
                if d_date:
                    label += " %s" % d_date
                if d_place:
                    label += " %s" % d_place

        if self.increlname and self.center_person != person:
            # display relationship info
            if self.advrelinfo:
                (relationship, _ga, _gb) = self.rel_calc.get_one_relationship(
                    self._db,
                    self.center_person,
                    person,
                    extra_info=True,
                    olocale=self._locale,
                )
                if relationship:
                    label += "%s(%s Ga=%d Gb=%d)" % (
                        line_delimiter,
                        relationship,
                        _ga,
                        _gb,
                    )
            else:
                relationship = self.rel_calc.get_one_relationship(
                    self._db, self.center_person, person, olocale=self._locale
                )
                if relationship:
                    label += "%s(%s)" % (line_delimiter, relationship)

        if self.occupation > 0:
            event_refs = person.get_primary_event_ref_list()
            events = [
                event
                for event in [
                    self._db.get_event_from_handle(ref.ref) for ref in event_refs
                ]
                if event.get_type() == EventType(EventType.OCCUPATION)
            ]
            if len(events) > 0:
                events.sort(key=lambda x: x.get_date_object())
                if self.occupation == 1:
                    occupation = events[-1].get_description()
                    if occupation:
                        label += "%s(%s)" % (line_delimiter, occupation)
                elif self.occupation == 2:
                    for evt in events:
                        date = self.get_date_string(evt)
                        place = self.get_place_string(evt)
                        desc = evt.get_description()
                        if not date and not desc and not place:
                            continue
                        label += "%s(" % line_delimiter
                        if date:
                            label += "%s" % date
                            if desc:
                                label += " "
                        if desc:
                            label += "%s" % desc
                        if place:
                            if date or desc:
                                label += self._(", ")  # Arabic OK
                            label += "%s" % place
                        label += ")"

        # see if we have a table that needs to be terminated
        if self.use_html_output:
            label += "</TD></TR></TABLE>"
            return label
        else:
            # non html label is enclosed by "" so escape other "
            return label.replace('"', '\\"')

    def get_event_strings(self, person):
        "returns tuple of birth/christening and death/burying date strings"

        birth_date = birth_place = death_date = death_place = ""
        birth_type = death_type = ""

        birth_event = get_birth_or_fallback(self._db, person)
        if birth_event:
            birth_type = birth_event.type
            birth_date = self.get_date_string(birth_event)
            birth_place = self.get_place_string(birth_event)

        death_event = get_death_or_fallback(self._db, person)
        if death_event:
            death_type = death_event.type
            death_date = self.get_date_string(death_event)
            death_place = self.get_place_string(death_event)

        return (
            birth_date,
            death_date,
            birth_place,
            death_place,
            birth_type,
            death_type,
        )

    def get_date_string(self, event):
        """
        return date string for an event label.

        Based on the data availability and preferences, we select one
        of the following for a given event:
            year only
            complete date
            empty string
        """
        if event and event.get_date_object() is not None:
            event_date = event.get_date_object()
            if event_date.get_year_valid():
                if self.event_choice in [4, 5]:
                    return self.get_date(Date(event_date.get_year()))  # localized year
                elif self.event_choice in [1, 2, 3, 7]:
                    return self.get_date(event_date)
        return ""

    def get_place_string(self, event):
        """
        return place string for an event label.

        Based on the data availability and preferences, we select one
        of the following for a given event:
            place name
            empty string
        """
        if event and self.event_choice in [2, 3, 5, 6, 7]:
            place = _pd.display_event(self._db, event)
            return html.escape(place)
        return ""

    def get_date(self, date):
        """return a formatted date"""
        return html.escape(self._get_date(date))


# ------------------------------------------------------------------------
#
# RelGraphOptions class
#
# ------------------------------------------------------------------------
class RelGraphOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__pid = None
        self.__filter = None
        self.__show_relships = None
        self.__show_ga_gb = None
        self.__include_images = None
        self.__image_on_side = None
        self.__db = dbase
        self._nf = None
        self.event_choice = None
        MenuReportOptions.__init__(self, name, dbase)

    def add_menu_options(self, menu):
        ################################
        category_name = _("Report Options")
        add_option = partial(menu.add_option, category_name)
        ################################

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(_("Determines what people are included in the graph"))
        add_option("filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        arrow = EnumeratedListOption(_("Arrowhead direction"), "d")
        for i in range(0, len(_ARROWS)):
            arrow.add_item(_ARROWS[i]["value"], _ARROWS[i]["name"])
        arrow.set_help(_("Choose the direction that the arrows point."))
        add_option("arrow", arrow)

        color = EnumeratedListOption(_("Graph coloring"), "filled")
        for i in range(0, len(_COLORS)):
            color.add_item(_COLORS[i]["value"], _COLORS[i]["name"])
        color.set_help(
            _(
                "Males will be shown with blue, females "
                "with red.  If the sex of an individual "
                "is unknown it will be shown with gray."
            )
        )
        add_option("color", color)

        # see bug report #2180
        roundedcorners = BooleanOption(_("Use rounded corners"), False)
        roundedcorners.set_help(
            _("Use rounded corners to differentiate " "between women and men.")
        )
        add_option("useroundedcorners", roundedcorners)

        # see bug report #11112
        hexagons = BooleanOption(_("Use hexagons"), False)
        hexagons.set_help(
            _("Use hexagons to differentiate " "those of unknown gender.")
        )
        add_option("usehexagons", hexagons)

        stdoptions.add_gramps_id_option(menu, category_name, ownline=True)

        ################################
        category_name = _("Report Options (2)")
        add_option = partial(menu.add_option, category_name)
        ################################

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        self.__update_filters()

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

        ################################
        add_option = partial(menu.add_option, _("Include"))
        ################################

        self.event_choice = EnumeratedListOption(_("Dates and/or Places"), 0)
        self.event_choice.add_item(0, _("Do not include any dates or places"))
        self.event_choice.add_item(
            1, _("Include (birth, marriage, death) " "dates, but no places")
        )
        self.event_choice.add_item(
            2, _("Include (birth, marriage, death) " "dates, and places")
        )
        self.event_choice.add_item(
            3, _("Include (birth, marriage, death) " "dates, and places if no dates")
        )
        self.event_choice.add_item(
            4, _("Include (birth, marriage, death) " "years, but no places")
        )
        self.event_choice.add_item(
            5, _("Include (birth, marriage, death) " "years, and places")
        )
        self.event_choice.add_item(
            6, _("Include (birth, marriage, death) " "places, but no dates")
        )
        self.event_choice.add_item(
            7, _("Include (birth, marriage, death) " "dates and places on same line")
        )
        self.event_choice.set_help(_("Whether to include dates and/or places"))
        add_option("event_choice", self.event_choice)

        show_family_leaves = BooleanOption(_("Show all family nodes"), True)
        show_family_leaves.set_help(
            _(
                "Show family nodes even if the output "
                "contains only one member of the family."
            )
        )
        add_option("show_family_leaves", show_family_leaves)

        url = BooleanOption(_("Include URLs"), False)
        url.set_help(
            _(
                "Include a URL in each graph node so "
                "that PDF and imagemap files can be "
                "generated that contain active links "
                "to the files generated by the 'Narrated "
                "Web Site' report."
            )
        )
        add_option("url", url)

        self.__show_relships = BooleanOption(
            _("Include relationship to center person"), False
        )
        self.__show_relships.set_help(
            _("Whether to show every person's " "relationship to the center person")
        )
        add_option("increlname", self.__show_relships)
        self.__show_relships.connect("value-changed", self.__show_relships_changed)

        self.__include_images = BooleanOption(
            _("Include thumbnail images of people"), False
        )
        self.__include_images.set_help(_("Whether to include thumbnails of people."))
        add_option("includeImages", self.__include_images)
        self.__include_images.connect("value-changed", self.__image_changed)

        self.__image_on_side = EnumeratedListOption(_("Thumbnail Location"), 0)
        self.__image_on_side.add_item(0, _("Above the name"))
        self.__image_on_side.add_item(1, _("Beside the name"))
        self.__image_on_side.set_help(
            _("Where the thumbnail image should appear relative to the name")
        )
        add_option("imageOnTheSide", self.__image_on_side)

        # occupation = BooleanOption(_("Include occupation"), False)
        occupation = EnumeratedListOption(_("Include occupation"), 0)
        occupation.add_item(0, _("Do not include any occupation"))
        occupation.add_item(1, _("Include description " "of most recent occupation"))
        occupation.add_item(
            2, _("Include date, description and place " "of all occupations")
        )
        occupation.set_help(_("Whether to include the last occupation"))
        add_option("occupation", occupation)

        if __debug__:
            self.__show_ga_gb = BooleanOption(
                _("Include relationship " "debugging numbers also"), False
            )
            self.__show_ga_gb.set_help(
                _(
                    "Whether to include 'Ga' and 'Gb' "
                    "also, to debug the relationship "
                    "calculator"
                )
            )
            add_option("advrelinfo", self.__show_ga_gb)

        ################################
        add_option = partial(menu.add_option, _("Graph Style"))
        ################################

        color_males = ColorOption(_("Males"), "#e0e0ff")
        color_males.set_help(_("The color to use to display men."))
        add_option("colormales", color_males)

        color_females = ColorOption(_("Females"), "#ffe0e0")
        color_females.set_help(_("The color to use to display women."))
        add_option("colorfemales", color_females)

        color_other = ColorOption(_("Other"), "#94ef9e")
        color_other.set_help(
            _("The color to use to display people who are " "neither men nor women.")
        )
        add_option("colorother", color_other)

        color_unknown = ColorOption(_("Unknown"), "#e0e0e0")
        color_unknown.set_help(_("The color to use when the gender is unknown."))
        add_option("colorunknown", color_unknown)

        color_family = ColorOption(_("Families"), "#ffffe0")
        color_family.set_help(_("The color to use to display families."))
        add_option("colorfamilies", color_family)

        dashed = BooleanOption(
            _("Indicate non-birth relationships with dotted lines"), True
        )
        dashed.set_help(
            _("Non-birth relationships will show up " "as dotted lines in the graph.")
        )
        add_option("dashed", dashed)

        showfamily = BooleanOption(_("Show family nodes"), True)
        showfamily.set_help(
            _("Families will show up as ellipses, linked " "to parents and children.")
        )
        add_option("showfamily", showfamily)

        use_subgraphs = EnumeratedListOption(_("Parent grouping"), 0)
        use_subgraphs.add_item(0, _("Normal"))
        use_subgraphs.add_item(1, _("Parents together"))
        use_subgraphs.add_item(2, _("Parents offset"))
        use_subgraphs.set_help(
            _(
                "In the 'Normal' setting parents will be located to keep most "
                "lines short.\n"
                "The 'Parents together' setting can help position "
                "spouses next to each other, but with non-trivial graphs will "
                "result in longer lines and larger graphs.\n"
                "The Parents offset setting will also try to put spouses near "
                "each other, however they will be offset from each other.  This "
                "will tend to make graphs with many people in a generation more "
                "square."
            )
        )
        add_option("usesubgraphs", use_subgraphs)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        nfv = self._nf.get_value()
        filter_list = utils.get_person_filters(
            person, include_single=False, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        if self.__show_relships and self.__show_relships.get_value():
            self.__pid.set_available(True)
        filter_value = self.__filter.get_value()
        if filter_value == 0:  # "Entire Database" (as "include_single=False")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def __image_changed(self):
        """
        Handle thumbnail change. If the image is not to be included, make the
        image location option unavailable.
        """
        self.__image_on_side.set_available(self.__include_images.get_value())

    def __show_relships_changed(self):
        """
        Enable/disable menu items if relationships are required
        """
        if self.__show_ga_gb:
            self.__show_ga_gb.set_available(self.__show_relships.get_value())
        self.__filter_changed()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Package providing filter rules for Gramps.
"""

from ._searchfathername import SearchFatherName
from ._searchmothername import SearchMotherName
from ._searchchildname import SearchChildName
from ._regexpfathername import RegExpFatherName
from ._regexpmothername import RegExpMotherName
from ._regexpchildname import RegExpChildName

from ._hasreltype import HasRelType
from ._allfamilies import AllFamilies
from ._hasgallery import HasGallery
from ._hasidof import HasIdOf
from ._haslds import HasLDS
from ._regexpidof import RegExpIdOf
from ._hasnote import HasNote
from ._hasnoteregexp import HasNoteRegexp
from ._hasnotematchingsubstringof import HasNoteMatchingSubstringOf
from ._hassourcecount import HasSourceCount
from ._hassourceof import HasSourceOf
from ._hasreferencecountof import HasReferenceCountOf
from ._hascitation import HasCitation
from ._familyprivate import FamilyPrivate
from ._hasattribute import HasAttribute
from ._hasevent import HasEvent
from ._isbookmarked import IsBookmarked
from ._matchesfilter import MatchesFilter
from ._matchessourceconfidence import MatchesSourceConfidence
from ._fatherhasnameof import FatherHasNameOf
from ._fatherhasidof import FatherHasIdOf
from ._motherhasnameof import MotherHasNameOf
from ._motherhasidof import MotherHasIdOf
from ._childhasnameof import ChildHasNameOf
from ._childhasidof import ChildHasIdOf
from ._changedsince import ChangedSince
from ._hastag import HasTag
from ._hastwins import HasTwins
from ._isancestorof import IsAncestorOf
from ._isdescendantof import IsDescendantOf

editor_rule_list = [
    AllFamilies,
    HasRelType,
    HasGallery,
    HasIdOf,
    HasLDS,
    HasNote,
    RegExpIdOf,
    HasNoteRegexp,
    HasReferenceCountOf,
    HasSourceCount,
    HasSourceOf,
    HasCitation,
    FamilyPrivate,
    HasEvent,
    HasAttribute,
    IsBookmarked,
    MatchesFilter,
    MatchesSourceConfidence,
    FatherHasNameOf,
    FatherHasIdOf,
    MotherHasNameOf,
    MotherHasIdOf,
    ChildHasNameOf,
    ChildHasIdOf,
    ChangedSince,
    HasTag,
    HasTwins,
    IsAncestorOf,
    IsDescendantOf,
]



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2013 Heinz Brinker <heinzbrinker@yahoo.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository.GLib import markup_escape_text

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import EventType, EventRoleType
from gramps.gen.plug import Gramplet
from gramps.gui.widgets import Photo
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.datehandler import get_date
from gramps.gen.utils.file import media_path_full
from gramps.gen.const import COLON, GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class PersonDetails(Gramplet):
    """
    Displays details for a person.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.uistate.connect("nameformat-changed", self.update)
        self.uistate.connect("placeformat-changed", self.update)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.Box()
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.photo = Photo(self.uistate.screen_height() < 1000)
        self.photo.show()
        self.name = Gtk.Label(halign=Gtk.Align.START)
        self.name.set_selectable(True)
        vbox.pack_start(self.name, fill=True, expand=False, padding=7)
        self.grid = Gtk.Grid(orientation=Gtk.Orientation.VERTICAL)
        self.grid.set_column_spacing(10)
        vbox.pack_start(self.grid, fill=True, expand=False, padding=5)
        vbox.show_all()
        self.top.pack_start(self.photo, fill=True, expand=False, padding=5)
        self.top.pack_start(vbox, fill=True, expand=True, padding=10)
        return self.top

    def add_row(self, title, value):
        """
        Add a row to the table.
        """
        label = Gtk.Label(
            label=title + COLON, halign=Gtk.Align.END, valign=Gtk.Align.START
        )
        label.set_selectable(True)
        label.show()
        value = Gtk.Label(label=value, halign=Gtk.Align.START)
        value.set_selectable(True)
        value.show()
        self.grid.add(label)
        self.grid.attach_next_to(value, label, Gtk.PositionType.RIGHT, 1, 1)

    def clear_grid(self):
        """
        Remove all the rows from the grid.
        """
        list(map(self.grid.remove, self.grid.get_children()))

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        """
        Determine if a person has_data by checking:

            1. has a birth, baptism, death, or burial event; OR
            2. has a father; OR
            3. has a mother
        """
        active_handle = self.get_active("Person")
        has_data = False
        if active_handle:
            active_person = self.dbstate.db.get_person_from_handle(active_handle)
            if active_person:
                for event_type in [
                    EventType(EventType.BIRTH),
                    EventType(EventType.BAPTISM),
                    EventType(EventType.DEATH),
                    EventType(EventType.BURIAL),
                ]:
                    event = self.get_event(active_person, event_type)
                    if event:
                        has_data = True
                        break
                if not has_data:
                    family_handle = active_person.get_main_parents_family_handle()
                    if family_handle:
                        family = self.dbstate.db.get_family_from_handle(family_handle)
                        handle = family.get_father_handle()
                        if handle:
                            if self.dbstate.db.get_person_from_handle(handle):
                                has_data = True
                            else:
                                handle = family.get_mother_handle()
                                if handle:
                                    if self.dbstate.db.get_person_from_handle(handle):
                                        has_data = True
        self.set_has_data(has_data)

    def main(self):  # return false finishes
        self.display_empty()
        active_handle = self.get_active("Person")
        if active_handle:
            active_person = self.dbstate.db.get_person_from_handle(active_handle)
            self.top.hide()
            if active_person:
                self.display_person(active_person)
            self.top.show()
        self.update_has_data()

    def display_person(self, active_person):
        """
        Display details of the active person.
        """
        self.load_person_image(active_person)
        self.name.set_markup(
            "<span size='large' weight='bold'>%s</span>"
            % markup_escape_text(name_displayer.display(active_person), -1)
        )
        self.clear_grid()
        self.display_alternate_names(active_person)
        self.display_parents(active_person)
        self.display_separator()
        self.display_type(active_person, EventType(EventType.BIRTH))
        self.display_type(active_person, EventType(EventType.BAPTISM))
        self.display_type(active_person, EventType(EventType.DEATH))
        self.display_type(active_person, EventType(EventType.BURIAL))
        self.display_separator()
        self.display_attribute(active_person, _("Occupation"))
        self.display_attribute(active_person, _("Title"))
        self.display_attribute(active_person, _("Religion"))

    def display_empty(self):
        """
        Display empty details when no person is selected.
        """
        self.photo.set_image(None)
        self.photo.set_uistate(None, None)
        self.name.set_text(_("No active person"))
        self.clear_grid()

    def display_separator(self):
        """
        Display an empty row to separate groupd of entries.
        """
        label = Gtk.Label()
        label.set_markup("<span font='sans 4'> </span>")
        label.set_selectable(True)
        label.show()
        self.grid.add(label)

    def display_alternate_names(self, active_person):
        """
        Display other names of the person
        """
        try:
            nlist = active_person.get_alternate_names()
            if len(nlist) > 0:
                for altname in nlist:
                    name_type = str(altname.get_type())
                    text = name_displayer.display_name(altname)
                    self.add_row(name_type, text)
                self.display_separator()
        except:
            pass

    def display_parents(self, active_person):
        """
        Display the parents of the active person.
        """
        family_handle = active_person.get_main_parents_family_handle()
        if family_handle:
            family = self.dbstate.db.get_family_from_handle(family_handle)
            handle = family.get_father_handle()
            if handle:
                father = self.dbstate.db.get_person_from_handle(handle)
                father_name = name_displayer.display(father)
            else:
                father_name = _("Unknown")
            handle = family.get_mother_handle()
            if handle:
                mother = self.dbstate.db.get_person_from_handle(handle)
                mother_name = name_displayer.display(mother)
            else:
                mother_name = _("Unknown")
        else:
            father_name = _("Unknown")
            mother_name = _("Unknown")

        self.add_row(_("Father"), father_name)
        self.add_row(_("Mother"), mother_name)

    def display_attribute(self, active_person, attr_key):
        """
        Display an attribute row.
        """
        values = []
        for attr in active_person.get_attribute_list():
            if attr.get_type() == attr_key:
                values.append(attr.get_value())
        if values:
            # Translators: needed for Arabic, ignore otherwise
            self.add_row(attr_key, _(", ").join(values))

    def display_type(self, active_person, event_type):
        """
        Display an event type row.
        """
        event = self.get_event(active_person, event_type)
        if event:
            self.add_row(str(event_type), self.format_event(event))

    def get_event(self, person, event_type):
        """
        Return an event of the given type.
        """
        for event_ref in person.get_event_ref_list():
            if int(event_ref.get_role()) == EventRoleType.PRIMARY:
                event = self.dbstate.db.get_event_from_handle(event_ref.ref)
                if event.get_type() == event_type:
                    return event
        return None

    def format_event(self, event):
        """
        Format the event for display.
        """
        date = get_date(event)
        handle = event.get_place_handle()
        if handle:
            place = place_displayer.display_event(self.dbstate.db, event)
            retval = _("%(date)s - %(place)s.") % {"date": date, "place": place}
        else:
            retval = _("%(date)s.") % dict(date=date)
        return retval

    def load_person_image(self, person):
        """
        Load the primary image if it exists.
        """
        media_list = person.get_media_list()
        if media_list:
            media_ref = media_list[0]
            object_handle = media_ref.get_reference_handle()
            obj = self.dbstate.db.get_media_from_handle(object_handle)
            full_path = media_path_full(self.dbstate.db, obj.get_path())
            mime_type = obj.get_mime_type()
            if mime_type and mime_type.startswith("image"):
                self.photo.set_image(full_path, mime_type, media_ref.get_rectangle())
                self.photo.set_uistate(self.uistate, object_handle)
            else:
                self.photo.set_image(None)
                self.photo.set_uistate(None, None)
        else:
            self.photo.set_image(None)
            self.photo.set_uistate(None, None)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2007  Donald N. Allingham
# Copyright (C) 2013       Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import os
import time
import logging
from xml.parsers.expat import ParserCreate, ExpatError

try:
    import fcntl

    USE_LOCK = True
except ImportError:
    USE_LOCK = False

from .const import USER_DATA, GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
GRAMPS_FILENAME = os.path.join(USER_DATA, "recent-files-gramps.xml")
MAX_GRAMPS_ITEMS = 10


# -------------------------------------------------------------------------
#
# RecentItem
#
# -------------------------------------------------------------------------
class RecentItem:
    """
    Interface to a single Gramps recent-items item
    """

    def __init__(self, p="", n="", t=0):
        self.path = p
        self.name = n
        self.time = t

    def set_path(self, val):
        """
        Set the file path.
        """
        self.path = val

    def get_path(self):
        """
        Get the file path.
        """
        return self.path

    def set_name(self, val):
        """
        Set the file name.
        """
        self.name = val

    def get_name(self):
        """
        Get the file name.
        """
        return self.name

    def set_time(self, val):
        """
        Set the file timestamp.
        """
        self.time = int(val)

    def get_time(self):
        """
        Get the file timestamp.
        """
        return self.time

    def __eq__(self, other_item):
        return self.time == other_item.time

    def __ne__(self, other_item):
        return self.time != other_item.time

    def __lt__(self, other_item):
        return self.time < other_item.time

    def __gt__(self, other_item):
        return self.time > other_item.time

    def __le__(self, other_item):
        return self.time <= other_item.time

    def __ge__(self, other_item):
        return self.time >= other_item.time


# -------------------------------------------------------------------------
#
# RecentFiles
#
# -------------------------------------------------------------------------
class RecentFiles:
    """
    Interface to a RecentFiles collection
    """

    def __init__(self):
        gramps_parser = RecentParser()
        self.gramps_recent_files = gramps_parser.get()

    def add(self, item2add):
        """
        Add a file to the recent files list.
        """
        # First we need to walk the existing items to see
        # if our item is already there
        for item in self.gramps_recent_files:
            if item.get_path() == item2add.get_path():
                # Found it -- modify timestamp and add all groups
                # to the item's groups
                item.set_time(item2add.get_time())
                return
        # At this point we walked the items and not found one,
        # so simply inserting a new item in the beginning
        self.gramps_recent_files.insert(0, item2add)

    def rename_filename(self, filename, new_filename):
        """
        Rename a file in the recent files list.
        """
        for recent_file in self.gramps_recent_files:
            if recent_file.get_name() == filename:
                recent_file.set_name(new_filename)
                return

    def remove_filename(self, filename):
        """
        Remove a file from the recent files list.
        """
        for index, recent_file in enumerate(self.gramps_recent_files):
            if recent_file.get_name() == filename:
                self.gramps_recent_files.pop(index)
                return

    def check_if_recent(self, filename):
        """
        Check if a file is present in the recent files list.
        """
        return any(
            filename == recent_file.get_name()
            for recent_file in self.gramps_recent_files
        )

    def save(self):
        """
        Attempt saving into XML.
        The trick is not to fail under any circumstances.
        """
        fname = os.path.expanduser(GRAMPS_FILENAME)
        try:
            self.do_save(fname)
        except IOError as err:
            logging.warning(
                _("Unable to save list of recent DBs file {fname}: {error}").format(
                    fname=fname, error=err
                )
            )

    def do_save(self, fname):
        """
        Saves the current Gramps RecentFiles collection to the associated file.
        """
        with open(fname, "w", encoding="utf8") as xml_file:
            if USE_LOCK:
                fcntl.lockf(xml_file, fcntl.LOCK_EX)
            xml_file.write('<?xml version="1.0" encoding="utf-8"?>\n')
            xml_file.write("<RecentFiles>\n")
            for index, item in enumerate(self.gramps_recent_files):
                if index > MAX_GRAMPS_ITEMS:
                    break
                xml_file.write("  <RecentItem>\n")
                xml_file.write("    <Path><![CDATA[%s]]></Path>\n" % item.get_path())
                xml_file.write("    <Name><![CDATA[%s]]></Name>\n" % item.get_name())
                xml_file.write("    <Timestamp>%d</Timestamp>\n" % item.get_time())
                xml_file.write("  </RecentItem>\n")
            xml_file.write("</RecentFiles>\n")

        # all advisory locks on a file are released on close


# -------------------------------------------------------------------------
#
# RecentParser
#
# -------------------------------------------------------------------------
class RecentParser:
    """
    Parsing class for the RecentFiles collection.
    """

    def __init__(self):
        self.recent_files = []
        self.tlist = []
        self.item = None

        fname = os.path.expanduser(GRAMPS_FILENAME)
        if not os.path.exists(fname):
            return  # it's the first time gramps has ever been run

        try:
            with open(fname, "rb") as xml_file:
                if USE_LOCK:
                    fcntl.lockf(xml_file, fcntl.LOCK_SH)

                parser = ParserCreate()
                parser.StartElementHandler = self.start_element
                parser.EndElementHandler = self.end_element
                parser.CharacterDataHandler = self.characters
                parser.ParseFile(xml_file)
            # all advisory locks on a file are released on close
        except IOError as err:
            logging.warning(
                _("Unable to open list of recent DBs file {fname}: {error}").format(
                    fname=fname, error=err
                )
            )
        except ExpatError as err:
            logging.error(
                _(
                    "Error parsing list of recent DBs from file {fname}: "
                    "{error}.\nThis might indicate a damage to your files.\n"
                    "If you're sure there is no problem with other files, "
                    "delete it, and restart Gramps."
                ).format(fname=fname, error=err)
            )

    def get(self):
        """
        Return a list of recent files.
        """
        return self.recent_files

    def start_element(self, tag, attrs):
        """
        Handler for XML start element.
        """
        self.tlist = []
        if tag == "RecentItem":
            self.item = RecentItem()

    def end_element(self, tag):
        """
        Handler for XML end element.
        """
        text = "".join(self.tlist)

        if tag == "RecentItem":
            if os.path.isdir(self.item.get_path()):
                self.recent_files.append(self.item)
        elif tag == "Path":
            self.item.set_path(text)
        elif tag == "Name":
            self.item.set_name(text)
        elif tag == "Timestamp":
            self.item.set_time(int(text))

    def characters(self, data):
        """
        Handler for XML character data.
        """
        self.tlist.append(data)


# -------------------------------------------------------------------------
#
# Helper functions
#
# -------------------------------------------------------------------------
def recent_files(filename, name):
    """
    Add an entry to the Gramps recent items list.
    """
    the_time = int(time.time())
    gramps_rf = RecentFiles()
    gramps_item = RecentItem(p=filename, n=name, t=the_time)
    gramps_rf.add(gramps_item)
    gramps_rf.save()


def remove_filename(filename):
    """
    Remove an entry from the Gramps recent items list.
    """
    gramps_rf = RecentFiles()
    gramps_rf.remove_filename(filename)
    gramps_rf.save()


def rename_filename(filename, new_filename):
    """
    Rename an entry in the Gramps recent items list.
    """
    gramps_rf = RecentFiles()
    gramps_rf.rename_filename(filename, new_filename)
    gramps_rf.save()


def check_if_recent(filename):
    """
    Check if an entry is present in Gramps recent items list.
    """
    gramps_rf = RecentFiles()
    return gramps_rf.check_if_recent(filename)
